# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""fuzz_task tests."""
# pylint: disable=protected-access

import datetime
import os
import queue
import shutil
import tempfile
import threading
import time
import unittest

import mock
import parameterized
from pyfakefs import fake_filesystem_unittest
import six

from clusterfuzz import stacktraces
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.fuzzers.libFuzzer import \
    engine as libfuzzer_engine
from clusterfuzz._internal.bot.tasks import fuzz_task
from clusterfuzz._internal.bot.untrusted_runner import file_host
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.chrome import crash_uploader
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.metrics import monitor
from clusterfuzz._internal.metrics import monitoring_metrics
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from clusterfuzz._internal.tests.test_libs import untrusted_runner_helpers
from clusterfuzz.fuzz import engine


class TrackFuzzerRunResultTest(unittest.TestCase):
  """Test _track_fuzzer_run_result."""

  def setUp(self):
    monitor.metrics_store().reset_for_testing()

  def test_fuzzer_run_result(self):
    """Ensure _track_fuzzer_run_result set the right metrics."""
    fuzz_task._track_fuzzer_run_result('name', 10, 100, 2)
    fuzz_task._track_fuzzer_run_result('name', 100, 200, 2)
    fuzz_task._track_fuzzer_run_result('name', 1000, 2000, 2)
    fuzz_task._track_fuzzer_run_result('name', 1000, 500, 0)
    fuzz_task._track_fuzzer_run_result('name', 0, 1000, -1)
    fuzz_task._track_fuzzer_run_result('name', 0, 0, 2)

    self.assertEqual(
        4,
        monitoring_metrics.FUZZER_RETURN_CODE_COUNT.get({
            'fuzzer': 'name',
            'return_code': 2
        }))
    self.assertEqual(
        1,
        monitoring_metrics.FUZZER_RETURN_CODE_COUNT.get({
            'fuzzer': 'name',
            'return_code': 0
        }))
    self.assertEqual(
        1,
        monitoring_metrics.FUZZER_RETURN_CODE_COUNT.get({
            'fuzzer': 'name',
            'return_code': -1
        }))

    testcase_count_ratio = (
        monitoring_metrics.FUZZER_TESTCASE_COUNT_RATIO.get({
            'fuzzer': 'name'
        }))
    self.assertEqual(3.1, testcase_count_ratio.sum)
    self.assertEqual(5, testcase_count_ratio.count)

    expected_buckets = [0 for _ in range(22)]
    expected_buckets[1] = 1
    expected_buckets[3] = 1
    expected_buckets[11] = 2
    expected_buckets[21] = 1
    self.assertListEqual(expected_buckets, testcase_count_ratio.buckets)


class TrackBuildRunResultTest(unittest.TestCase):
  """Test _track_build_run_result."""

  def setUp(self):
    monitor.metrics_store().reset_for_testing()

  def test_build_run_result(self):
    """Ensure _track_build_run_result set the right metrics."""
    fuzz_task._track_build_run_result('name', 10000, True)
    fuzz_task._track_build_run_result('name', 10001, True)
    fuzz_task._track_build_run_result('name', 10002, False)

    self.assertEqual(
        2,
        monitoring_metrics.JOB_BAD_BUILD_COUNT.get({
            'job': 'name',
            'bad_build': True
        }))
    self.assertEqual(
        1,
        monitoring_metrics.JOB_BAD_BUILD_COUNT.get({
            'job': 'name',
            'bad_build': False
        }))


class TrackTestcaseRunResultTest(unittest.TestCase):
  """Test _track_testcase_run_result."""

  def setUp(self):
    monitor.metrics_store().reset_for_testing()

  def test_testcase_run_result(self):
    """Ensure _track_testcase_run_result sets the right metrics."""
    fuzz_task._track_testcase_run_result('fuzzer', 'job', 2, 5)
    fuzz_task._track_testcase_run_result('fuzzer', 'job', 5, 10)

    self.assertEqual(7,
                     monitoring_metrics.JOB_NEW_CRASH_COUNT.get({
                         'job': 'job'
                     }))
    self.assertEqual(
        15, monitoring_metrics.JOB_KNOWN_CRASH_COUNT.get({
            'job': 'job'
        }))
    self.assertEqual(
        7, monitoring_metrics.FUZZER_NEW_CRASH_COUNT.get({
            'fuzzer': 'fuzzer'
        }))
    self.assertEqual(
        15, monitoring_metrics.FUZZER_KNOWN_CRASH_COUNT.get({
            'fuzzer': 'fuzzer'
        }))


class TruncateFuzzerOutputTest(unittest.TestCase):
  """Truncate fuzzer output tests."""

  def test_no_truncation(self):
    """No truncation."""
    self.assertEqual('aaaa', fuzz_task.truncate_fuzzer_output('aaaa', 10))

  def test_truncation(self):
    """Truncate."""
    self.assertEqual(
        '123456\n...truncated...\n54321',
        fuzz_task.truncate_fuzzer_output(
            '123456xxxxxxxxxxxxxxxxxxxxxxxxxxx54321', 28))

  def test_error(self):
    """Error if limit is too low."""
    with self.assertRaises(AssertionError):
      self.assertEqual(
          '', fuzz_task.truncate_fuzzer_output('123456xxxxxx54321', 10))


class TrackFuzzTimeTest(unittest.TestCase):
  """Test _TrackFuzzTime."""

  def setUp(self):
    monitor.metrics_store().reset_for_testing()

  def _test(self, timeout):
    """Test helper."""
    time_module = helpers.MockTime()
    with fuzz_task._TrackFuzzTime('fuzzer', 'job', time_module) as tracker:
      time_module.advance(5)
      tracker.timeout = timeout

    fuzzer_total_time = monitoring_metrics.FUZZER_TOTAL_FUZZ_TIME.get({
        'fuzzer': 'fuzzer',
        'timeout': timeout
    })
    self.assertEqual(5, fuzzer_total_time)

  def test_success(self):
    """Test report metrics."""
    self._test(False)

  def test_timeout(self):
    """Test timeout."""
    self._test(True)


class GetFuzzerMetadataFromOutputTest(unittest.TestCase):
  """Test get_fuzzer_metadata_from_output."""

  def test_no_metadata(self):
    """Tests no metadata in output."""
    data = 'abc\ndef\n123123'
    self.assertDictEqual(fuzz_task.get_fuzzer_metadata_from_output(data), {})

    data = ''
    self.assertDictEqual(fuzz_task.get_fuzzer_metadata_from_output(data), {})

  def test_metadata(self):
    """Tests parsing of metadata."""
    data = ('abc\n'
            'def\n'
            'metadata:invalid: invalid\n'
            'metadat::invalid: invalid\n'
            'metadata::foo: bar\n'
            '123123\n'
            'metadata::blah: 1\n'
            'metadata::test:abcd\n'
            'metadata::test2:   def\n')
    self.assertDictEqual(
        fuzz_task.get_fuzzer_metadata_from_output(data), {
            'blah': '1',
            'test': 'abcd',
            'test2': 'def',
            'foo': 'bar'
        })


class GetRegressionTest(unittest.TestCase):
  """Test get_regression."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager.is_custom_binary'
    ])

  def test_one_time_crasher(self):
    """Test when one_time_crasher_flag is True."""
    self.mock.is_custom_binary.return_value = False
    self.assertEqual('NA', fuzz_task.get_regression(True))

  def test_custom_binary(self):
    """Test for custom binary."""
    self.mock.is_custom_binary.return_value = True
    self.assertEqual('NA', fuzz_task.get_regression(False))

  def test_reproducible_non_custom_binary(self):
    """Test for reproducible non-custom binary."""
    self.mock.is_custom_binary.return_value = False
    self.assertEqual('', fuzz_task.get_regression(False))


class GetFixedOrMinimizedKeyTest(unittest.TestCase):
  """Test get_fixed_or_minimized_key."""

  def test_one_time_crasher(self):
    """Test when one_time_crasher_flag is True."""
    self.assertEqual('NA', fuzz_task.get_fixed_or_minimized_key(True))

  def test_reproducible(self):
    """Test for reproducible."""
    self.assertEqual('', fuzz_task.get_fixed_or_minimized_key(False))


class CrashInitTest(fake_filesystem_unittest.TestCase):
  """Test Crash.__init__."""

  def setUp(self):
    """Setup for crash init test."""
    helpers.patch(self, [
        'clusterfuzz._internal.chrome.crash_uploader.FileMetadataInfo',
        'clusterfuzz._internal.bot.tasks.setup.archive_testcase_and_dependencies_in_gcs',
        'clusterfuzz._internal.crash_analysis.stack_parsing.stack_analyzer.get_crash_data',
        'clusterfuzz._internal.bot.testcase_manager.get_additional_command_line_flags',
        'clusterfuzz._internal.bot.testcase_manager.get_command_line_for_application',
        'clusterfuzz._internal.base.utils.get_crash_stacktrace_output',
        'clusterfuzz._internal.crash_analysis.crash_analyzer.ignore_stacktrace',
        'clusterfuzz._internal.crash_analysis.crash_analyzer.is_security_issue',
    ])
    helpers.patch_environ(self)
    test_utils.set_up_pyfakefs(self)

    self.mock.get_command_line_for_application.return_value = 'cmd'
    dummy_state = stacktraces.CrashInfo()
    dummy_state.crash_type = 'type'
    dummy_state.crash_address = 'address'
    dummy_state.crash_state = 'state'
    dummy_state.crash_stacktrace = 'orig_trace'
    dummy_state.frames = ['frame 1', 'frame 2']
    self.mock.get_crash_data.return_value = dummy_state
    self.mock.get_crash_stacktrace_output.return_value = 'trace'
    self.mock.archive_testcase_and_dependencies_in_gcs.return_value = (
        'fuzzed_key', True, 'absolute_path', 'archive_filename')

    environment.set_value('FILTER_FUNCTIONAL_BUGS', False)

    with open('/stack_file_path', 'w') as f:
      f.write('unsym')

  def test_error(self):
    """Test failing to reading stacktrace file."""
    crash = fuzz_task.Crash.from_testcase_manager_crash(
        testcase_manager.Crash('dir/path-http-name', 123, 11, ['res'], 'ges',
                               '/no_stack_file'))
    self.assertIsNone(crash)

  def _test_crash(self, should_be_ignored, security_flag):
    """Test crash."""
    self.mock.get_command_line_for_application.reset_mock()
    self.mock.get_crash_data.reset_mock()
    self.mock.get_crash_stacktrace_output.reset_mock()
    self.mock.is_security_issue.reset_mock()
    self.mock.ignore_stacktrace.reset_mock()

    self.mock.is_security_issue.return_value = security_flag
    self.mock.ignore_stacktrace.return_value = should_be_ignored

    crash = fuzz_task.Crash.from_testcase_manager_crash(
        testcase_manager.Crash('dir/path-http-name', 123, 11, ['res'], 'ges',
                               '/stack_file_path'))

    self.assertEqual('dir/path-http-name', crash.file_path)
    self.assertEqual(123, crash.crash_time)
    self.assertEqual(11, crash.return_code)
    self.assertListEqual(['res'], crash.resource_list)
    self.assertEqual('ges', crash.gestures)

    self.assertEqual('path-http-name', crash.filename)
    self.assertTrue(crash.http_flag)

    self.assertEqual('cmd', crash.application_command_line)
    self.mock.get_command_line_for_application.assert_called_once_with(
        'dir/path-http-name', needs_http=True)

    self.assertEqual('unsym', crash.unsymbolized_crash_stacktrace)

    self.assertEqual('type', crash.crash_type)
    self.assertEqual('address', crash.crash_address)
    self.assertEqual('state', crash.crash_state)
    self.assertListEqual(['frame 1', 'frame 2'], crash.crash_frames)
    self.mock.get_crash_data.assert_called_once_with('unsym')

    self.assertEqual('trace', crash.crash_stacktrace)
    self.mock.get_crash_stacktrace_output.assert_called_once_with(
        'cmd', 'orig_trace', 'unsym')

    self.assertEqual(security_flag, crash.security_flag)
    self.mock.is_security_issue.assert_called_once_with('unsym', 'type',
                                                        'address')

    self.assertEqual('type,state,%s' % security_flag, crash.key)

    self.assertEqual(should_be_ignored, crash.should_be_ignored)
    self.mock.ignore_stacktrace.assert_called_once_with('orig_trace')

    self.assertFalse(hasattr(crash, 'fuzzed_key'))
    return crash

  def _test_validity_and_get_functional_crash(self):
    """Test validity of different crashes and return functional crash."""
    security_crash = self._test_crash(
        should_be_ignored=False, security_flag=True)
    self.assertIsNone(security_crash.get_error())
    self.assertTrue(security_crash.is_valid())

    ignored_crash = self._test_crash(should_be_ignored=True, security_flag=True)
    self.assertIn('False crash', ignored_crash.get_error())
    self.assertFalse(ignored_crash.is_valid())

    functional_crash = self._test_crash(
        should_be_ignored=False, security_flag=False)
    return functional_crash

  def test_valid_functional_bug(self):
    """Test valid because of functional bug."""
    functional_crash = self._test_validity_and_get_functional_crash()

    self.assertIsNone(functional_crash.get_error())
    self.assertTrue(functional_crash.is_valid())

  def test_invalid_functional_bug(self):
    """Test invalid because of functional bug."""
    environment.set_value('FILTER_FUNCTIONAL_BUGS', True)
    functional_crash = self._test_validity_and_get_functional_crash()

    self.assertIn('Functional crash', functional_crash.get_error())
    self.assertFalse(functional_crash.is_valid())

  def test_hydrate_fuzzed_key(self):
    """Test hydrating fuzzed_key."""
    crash = self._test_crash(should_be_ignored=False, security_flag=True)
    self.assertFalse(crash.is_archived())
    self.assertIsNone(crash.get_error())
    self.assertTrue(crash.is_valid())

    crash.archive_testcase_in_blobstore()
    self.assertTrue(crash.is_archived())
    self.assertIsNone(crash.get_error())
    self.assertTrue(crash.is_valid())

    self.assertEqual('fuzzed_key', crash.fuzzed_key)
    self.assertTrue(crash.archived)
    self.assertEqual('absolute_path', crash.absolute_path)
    self.assertEqual('archive_filename', crash.archive_filename)

  def test_hydrate_fuzzed_key_failure(self):
    """Test fail to hydrate fuzzed_key."""
    self.mock.archive_testcase_and_dependencies_in_gcs.return_value = (None,
                                                                       False,
                                                                       None,
                                                                       None)

    crash = self._test_crash(should_be_ignored=False, security_flag=True)
    self.assertFalse(crash.is_archived())
    self.assertIsNone(crash.get_error())
    self.assertTrue(crash.is_valid())

    crash.archive_testcase_in_blobstore()
    self.assertTrue(crash.is_archived())
    self.assertIn('Unable to store testcase in blobstore', crash.get_error())
    self.assertFalse(crash.is_valid())

    self.assertIsNone(crash.fuzzed_key)
    self.assertFalse(crash.archived)
    self.assertIsNone(crash.absolute_path)
    self.assertIsNone(crash.archive_filename)

  def test_args_from_testcase_manager(self):
    """Test args from testcase_manager.Crash."""
    testcase_manager_crash = testcase_manager.Crash('path', 0, 0, [], [],
                                                    '/stack_file_path')
    self.mock.get_additional_command_line_flags.return_value = 'minimized'
    environment.set_value('APP_ARGS', 'app')

    crash = fuzz_task.Crash.from_testcase_manager_crash(testcase_manager_crash)
    self.assertEqual('app minimized', crash.arguments)


class CrashGroupTest(unittest.TestCase):
  """Test CrashGroup."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.bot.tasks.fuzz_task.find_main_crash',
        'clusterfuzz._internal.datastore.data_handler.find_testcase',
        'clusterfuzz._internal.datastore.data_handler.get_project_name',
    ])

    self.mock.get_project_name.return_value = 'some_project'
    self.crashes = [self._make_crash('g1'), self._make_crash('g2')]
    self.context = mock.MagicMock(
        test_timeout=99, fuzzer_name='test', fuzz_target=None)
    self.reproducible_testcase = self._make_testcase(
        project_name='some_project',
        bug_information='',
        one_time_crasher_flag=False)
    self.unreproducible_testcase = self._make_testcase(
        project_name='some_project',
        bug_information='',
        one_time_crasher_flag=True)

  def _make_crash(self, gestures):
    crash = mock.MagicMock(
        crash_type='type',
        crash_state='state',
        security_flag=True,
        file_path='file_path',
        http_flag=True,
        gestures=gestures)
    return crash

  def _make_testcase(self,
                     project_name,
                     bug_information,
                     one_time_crasher_flag,
                     timestamp=datetime.datetime.now()):
    """Make testcase."""
    testcase = data_types.Testcase()
    testcase.timestamp = timestamp
    testcase.one_time_crasher_flag = one_time_crasher_flag
    testcase.bug_information = bug_information
    testcase.project_name = project_name
    return testcase

  def test_no_existing_testcase(self):
    """is_new=True and should_create_testcase=True when there's no existing
        testcase."""
    self.mock.find_testcase.return_value = None
    self.mock.find_main_crash.return_value = self.crashes[0], True

    group = fuzz_task.CrashGroup(self.crashes, self.context)

    self.assertTrue(group.should_create_testcase())
    self.mock.find_main_crash.assert_called_once_with(
        self.crashes, 'test', 'test', self.context.test_timeout)

    self.assertIsNone(group.existing_testcase)
    self.assertEqual(self.crashes[0], group.main_crash)
    self.assertTrue(group.is_new())

  def test_has_existing_reproducible_testcase(self):
    """should_create_testcase=False when there's an existing reproducible
      testcase."""
    self.mock.find_testcase.return_value = self.reproducible_testcase
    self.mock.find_main_crash.return_value = (self.crashes[0], True)

    group = fuzz_task.CrashGroup(self.crashes, self.context)

    self.assertEqual(self.crashes[0].gestures, group.main_crash.gestures)
    self.mock.find_main_crash.assert_called_once_with(
        self.crashes, 'test', 'test', self.context.test_timeout)
    self.assertFalse(group.is_new())
    self.assertFalse(group.should_create_testcase())
    self.assertTrue(group.has_existing_reproducible_testcase())

  def test_reproducible_crash(self):
    """should_create_testcase=True when the group is reproducible."""
    self.mock.find_testcase.return_value = self.unreproducible_testcase
    self.mock.find_main_crash.return_value = (self.crashes[0], False)

    group = fuzz_task.CrashGroup(self.crashes, self.context)

    self.assertEqual(self.crashes[0].gestures, group.main_crash.gestures)
    self.mock.find_main_crash.assert_called_once_with(
        self.crashes, 'test', 'test', self.context.test_timeout)
    self.assertFalse(group.is_new())
    self.assertTrue(group.should_create_testcase())
    self.assertFalse(group.has_existing_reproducible_testcase())
    self.assertFalse(group.one_time_crasher_flag)

  def test_has_existing_unreproducible_testcase(self):
    """should_create_testcase=False when the unreproducible testcase already
    exists."""
    self.mock.find_testcase.return_value = self.unreproducible_testcase
    self.mock.find_main_crash.return_value = (self.crashes[0], True)

    group = fuzz_task.CrashGroup(self.crashes, self.context)

    self.assertFalse(group.should_create_testcase())

    self.assertEqual(self.crashes[0].gestures, group.main_crash.gestures)
    self.mock.find_main_crash.assert_called_once_with(
        self.crashes, 'test', 'test', self.context.test_timeout)
    self.assertFalse(group.is_new())
    self.assertFalse(group.has_existing_reproducible_testcase())
    self.assertTrue(group.one_time_crasher_flag)


class FindMainCrashTest(unittest.TestCase):
  """Test find_main_crash."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.bot.testcase_manager.test_for_reproducibility',
    ])
    self.crashes = [
        self._make_crash('g1'),
        self._make_crash('g2'),
        self._make_crash('g3'),
        self._make_crash('g4')
    ]
    self.reproducible_crashes = []

    # pylint: disable=unused-argument
    def test_for_repro(fuzzer_name,
                       full_fuzzer_name,
                       file_path,
                       state,
                       security_flag,
                       test_timeout,
                       http_flag,
                       gestures,
                       arguments=None):
      """Mock test_for_reproducibility."""
      for c in self.reproducible_crashes:
        if c.gestures == gestures:
          return True
      return False

    self.mock.test_for_reproducibility.side_effect = test_for_repro

  def _make_crash(self, gestures):
    crash = mock.MagicMock(
        file_path='file_path',
        crash_state='state',
        security_flag=True,
        test_timeout=999,
        gestures=gestures)
    return crash

  def test_reproducible_crash(self):
    """Find that the 2nd crash is reproducible."""
    for c in self.crashes:
      c.is_valid.return_value = True
    self.crashes[0].is_valid.return_value = False
    self.reproducible_crashes = [self.crashes[2]]

    self.assertEqual((self.crashes[2], False),
                     fuzz_task.find_main_crash(self.crashes, 'test', 'test',
                                               99))

    self.crashes[0].archive_testcase_in_blobstore.assert_called_once_with()
    self.crashes[1].archive_testcase_in_blobstore.assert_called_once_with()
    self.crashes[2].archive_testcase_in_blobstore.assert_called_once_with()
    self.crashes[3].archive_testcase_in_blobstore.assert_not_called()

    # Calls for self.crashes[1] and self.crashes[2].
    self.assertEqual(2, self.mock.test_for_reproducibility.call_count)

  def test_unreproducible_crash(self):
    """No reproducible crash. Find the first valid one."""
    for c in self.crashes:
      c.is_valid.return_value = True
    self.crashes[0].is_valid.return_value = False
    self.reproducible_crashes = []

    self.assertEqual((self.crashes[1], True),
                     fuzz_task.find_main_crash(self.crashes, 'test', 'test',
                                               99))

    for c in self.crashes:
      c.archive_testcase_in_blobstore.assert_called_once_with()

    # Calls for every crash except self.crashes[0] because it's invalid.
    self.assertEqual(
        len(self.crashes) - 1, self.mock.test_for_reproducibility.call_count)

  def test_no_valid_crash(self):
    """No valid crash."""
    for c in self.crashes:
      c.is_valid.return_value = False
    self.reproducible_crashes = []

    self.assertEqual((None, None),
                     fuzz_task.find_main_crash(self.crashes, 'test', 'test',
                                               99))

    for c in self.crashes:
      c.archive_testcase_in_blobstore.assert_called_once_with()

    self.assertEqual(0, self.mock.test_for_reproducibility.call_count)


@test_utils.with_cloud_emulators('datastore')
class ProcessCrashesTest(fake_filesystem_unittest.TestCase):
  """Test process_crashes."""

  def setUp(self):
    """Setup for process crashes test."""
    helpers.patch(self, [
        'clusterfuzz._internal.chrome.crash_uploader.get_symbolized_stack_bytes',
        'clusterfuzz._internal.bot.tasks.fuzz_task.get_unsymbolized_crash_stacktrace',
        'clusterfuzz._internal.bot.tasks.task_creation.create_tasks',
        'clusterfuzz._internal.bot.tasks.setup.archive_testcase_and_dependencies_in_gcs',
        'clusterfuzz._internal.crash_analysis.stack_parsing.stack_analyzer.get_crash_data',
        'clusterfuzz._internal.build_management.revisions.get_real_revision',
        'clusterfuzz._internal.bot.testcase_manager.get_command_line_for_application',
        'clusterfuzz._internal.bot.testcase_manager.test_for_reproducibility',
        'clusterfuzz._internal.base.utils.get_crash_stacktrace_output',
        'clusterfuzz._internal.crash_analysis.crash_analyzer.ignore_stacktrace',
        'clusterfuzz._internal.crash_analysis.crash_analyzer.is_security_issue',
        'clusterfuzz._internal.datastore.data_handler.get_issue_tracker_name',
        'clusterfuzz._internal.datastore.data_handler.get_project_name',
        'clusterfuzz._internal.google_cloud_utils.big_query.Client.insert',
        'clusterfuzz._internal.google_cloud_utils.big_query.get_api_client',
        'time.sleep', 'time.time'
    ])
    test_utils.set_up_pyfakefs(self)

    self.mock.time.return_value = 987

    self.mock.get_issue_tracker_name.return_value = 'some_issue_tracker'
    self.mock.get_project_name.return_value = 'some_project'
    self.mock.archive_testcase_and_dependencies_in_gcs.return_value = (
        'fuzzed_key', True, 'absolute_path', 'archive_filename')

  def _make_crash(self, trace, state='state'):
    """Make crash."""
    self.mock.get_real_revision.return_value = 'this.is.fake.ver'

    self.mock.get_command_line_for_application.return_value = 'cmd'
    dummy_state = stacktraces.CrashInfo()
    dummy_state.crash_type = 'type'
    dummy_state.crash_address = 'address'
    dummy_state.crash_state = state
    dummy_state.crash_stacktrace = 'orig_trace'
    dummy_state.crash_frames = ['frame 1', 'frame 2']
    self.mock.get_crash_data.return_value = dummy_state
    self.mock.get_symbolized_stack_bytes.return_value = b'f00df00d'
    self.mock.get_crash_stacktrace_output.return_value = trace
    self.mock.get_unsymbolized_crash_stacktrace.return_value = trace
    self.mock.is_security_issue.return_value = True
    self.mock.ignore_stacktrace.return_value = False

    with open('/stack_file_path', 'w') as f:
      f.write('unsym')

    crash = fuzz_task.Crash.from_testcase_manager_crash(
        testcase_manager.Crash('dir/path-http-name', 123, 11, ['res'], ['ges'],
                               '/stack_file_path'))
    return crash

  def test_existing_unreproducible_testcase(self):
    """Test existing unreproducible testcase."""
    crashes = [self._make_crash('c1'), self._make_crash('c2')]
    self.mock.test_for_reproducibility.return_value = False

    existing_testcase = data_types.Testcase()
    existing_testcase.crash_stacktrace = 'existing'
    existing_testcase.crash_type = crashes[0].crash_type
    existing_testcase.crash_state = crashes[0].crash_state
    existing_testcase.security_flag = crashes[0].security_flag
    existing_testcase.one_time_crasher_flag = True
    existing_testcase.job_type = 'existing_job'
    existing_testcase.timestamp = datetime.datetime.now()
    existing_testcase.project_name = 'some_project'
    existing_testcase.put()

    variant = data_types.TestcaseVariant()
    variant.status = data_types.TestcaseVariantStatus.UNREPRODUCIBLE
    variant.job_type = 'job'
    variant.testcase_id = existing_testcase.key.id()
    variant.put()

    new_crash_count, known_crash_count, groups = fuzz_task.process_crashes(
        crashes=crashes,
        context=fuzz_task.Context(
            project_name='some_project',
            bot_name='bot',
            job_type='job',
            fuzz_target=data_types.FuzzTarget(engine='engine', binary='binary'),
            redzone=111,
            disable_ubsan=True,
            platform_id='platform',
            crash_revision=1234,
            fuzzer_name='fuzzer',
            window_argument='win_args',
            fuzzer_metadata={},
            testcases_metadata={},
            timeout_multiplier=1,
            test_timeout=2,
            thread_wait_timeout=3,
            data_directory='/data'))
    self.assertEqual(0, new_crash_count)
    self.assertEqual(2, known_crash_count)

    self.assertEqual(1, len(groups))
    self.assertEqual(2, len(groups[0].crashes))
    self.assertFalse(groups[0].is_new())
    self.assertEqual(crashes[0].crash_type, groups[0].main_crash.crash_type)
    self.assertEqual(crashes[0].crash_state, groups[0].main_crash.crash_state)
    self.assertEqual(crashes[0].security_flag,
                     groups[0].main_crash.security_flag)

    testcases = list(data_types.Testcase.query())
    self.assertEqual(1, len(testcases))
    self.assertEqual('existing', testcases[0].crash_stacktrace)

    variant = data_handler.get_testcase_variant(existing_testcase.key.id(),
                                                'job')
    self.assertEqual(data_types.TestcaseVariantStatus.FLAKY, variant.status)
    self.assertEqual('fuzzed_key', variant.reproducer_key)
    self.assertEqual(1234, variant.revision)
    self.assertEqual('type', variant.crash_type)
    self.assertEqual('state', variant.crash_state)
    self.assertEqual(True, variant.security_flag)
    self.assertEqual(True, variant.is_similar)

  @parameterized.parameterized.expand(['some_project', 'chromium'])
  def test_create_many_groups(self, project_name):
    """Test creating many groups."""
    self.mock.get_project_name.return_value = project_name

    self.mock.insert.return_value = {'insertErrors': [{'index': 0}]}

    # TODO(metzman): Add a separate test for strategies.
    r2_stacktrace = ('r2\ncf::fuzzing_strategies: value_profile\n')

    crashes = [
        self._make_crash('r1', state='reproducible1'),
        self._make_crash(r2_stacktrace, state='reproducible1'),
        self._make_crash('r3', state='reproducible1'),
        self._make_crash('r4', state='reproducible2'),
        self._make_crash('u1', state='unreproducible1'),
        self._make_crash('u2', state='unreproducible2'),
        self._make_crash('u3', state='unreproducible2'),
        self._make_crash('u4', state='unreproducible3')
    ]

    self.mock.test_for_reproducibility.side_effect = [
        False,  # For r1. It returns False. So, r1 is demoted.
        True,  # For r2. It returns True. So, r2 becomes primary for its group.
        True,  # For r4.
        False,  # For u1.
        False,  # For u2.
        False,  # For u3.
        False
    ]  # For u4.

    new_crash_count, known_crash_count, groups = fuzz_task.process_crashes(
        crashes=crashes,
        context=fuzz_task.Context(
            project_name=project_name,
            bot_name='bot',
            job_type='job',
            fuzz_target=data_types.FuzzTarget(engine='engine', binary='binary'),
            redzone=111,
            disable_ubsan=False,
            platform_id='platform',
            crash_revision=1234,
            fuzzer_name='fuzzer',
            window_argument='win_args',
            fuzzer_metadata={},
            testcases_metadata={},
            timeout_multiplier=1,
            test_timeout=2,
            thread_wait_timeout=3,
            data_directory='/data'))
    self.assertEqual(5, new_crash_count)
    self.assertEqual(3, known_crash_count)

    self.assertEqual(5, len(groups))
    self.assertEqual([
        'reproducible1', 'reproducible2', 'unreproducible1', 'unreproducible2',
        'unreproducible3'
    ], [group.main_crash.crash_state for group in groups])
    self.assertEqual([True, True, True, True, True],
                     [group.is_new() for group in groups])
    self.assertEqual([3, 1, 1, 2, 1], [len(group.crashes) for group in groups])

    testcases = list(data_types.Testcase.query())
    self.assertEqual(5, len(testcases))
    self.assertSetEqual(
        set([r2_stacktrace, 'r4', 'u1', 'u2', 'u4']),
        set(t.crash_stacktrace for t in testcases))

    self.assertSetEqual(
        set([
            '{"fuzzing_strategies": ["value_profile"]}', None, None, None, None
        ]), set(t.additional_metadata for t in testcases))

    # r2 is a reproducible crash, so r3 doesn't
    # invoke archive_testcase_in_blobstore. Therefore, the
    # archive_testcase_in_blobstore is called `len(crashes) - 1`.
    self.assertEqual(
        len(crashes) - 1,
        self.mock.archive_testcase_and_dependencies_in_gcs.call_count)

    # Check only the desired testcases were saved.
    actual_crash_infos = [group.main_crash.crash_info for group in groups]
    if project_name != 'chromium':
      expected_crash_infos = [None] * len(actual_crash_infos)
    else:
      expected_saved_crash_info = crash_uploader.CrashReportInfo(
          product='Chrome_' + environment.platform().lower().capitalize(),
          version='this.is.fake.ver',
          serialized_crash_stack_frames=b'f00df00d')
      expected_crash_infos = [
          expected_saved_crash_info,  # r2 is main crash for group r1,r2,r3
          expected_saved_crash_info,  # r4 is main crash for its own group
          None,  # u1 is not reproducible
          None,  # u2, u3 are not reproducible
          None,  # u4 is not reproducible
      ]

    self.assertEqual(len(expected_crash_infos), len(actual_crash_infos))
    for expected, actual in zip(expected_crash_infos, actual_crash_infos):
      if not expected:
        self.assertIsNone(actual)
        continue

      self.assertEqual(expected.product, actual.product)
      self.assertEqual(expected.version, actual.version)
      self.assertEqual(expected.serialized_crash_stack_frames,
                       actual.serialized_crash_stack_frames)

    def _make_big_query_json(crash, reproducible_flag, new_flag, testcase_id):
      return {
          'crash_type': crash.crash_type,
          'crash_state': crash.crash_state,
          'created_at': 987,
          'platform': 'platform',
          'crash_time_in_ms': int(crash.crash_time * 1000),
          'parent_fuzzer_name': 'engine',
          'fuzzer_name': 'engine_binary',
          'job_type': 'job',
          'security_flag': crash.security_flag,
          'reproducible_flag': reproducible_flag,
          'revision': '1234',
          'new_flag': new_flag,
          'project': project_name,
          'testcase_id': testcase_id
      }

    def _get_testcase_id(crash):
      rows = list(
          data_types.Testcase.query(
              data_types.Testcase.crash_type == crash.crash_type,
              data_types.Testcase.crash_state == crash.crash_state,
              data_types.Testcase.security_flag == crash.security_flag))
      if not rows:
        return None
      return str(rows[0].key.id())

    # Calls to write 5 groups of crashes to BigQuery.
    self.assertEqual(5, self.mock.insert.call_count)
    self.mock.insert.assert_has_calls([
        mock.call(mock.ANY, [
            big_query.Insert(
                _make_big_query_json(crashes[0], True, False, None),
                '%s:bot:987:0' % crashes[0].key),
            big_query.Insert(
                _make_big_query_json(crashes[1], True, True,
                                     _get_testcase_id(crashes[1])),
                '%s:bot:987:1' % crashes[0].key),
            big_query.Insert(
                _make_big_query_json(crashes[2], True, False, None),
                '%s:bot:987:2' % crashes[0].key)
        ]),
        mock.call(mock.ANY, [
            big_query.Insert(
                _make_big_query_json(crashes[3], True, True,
                                     _get_testcase_id(crashes[3])),
                '%s:bot:987:0' % crashes[3].key)
        ]),
        mock.call(mock.ANY, [
            big_query.Insert(
                _make_big_query_json(crashes[4], False, True,
                                     _get_testcase_id(crashes[4])),
                '%s:bot:987:0' % crashes[4].key)
        ]),
        mock.call(mock.ANY, [
            big_query.Insert(
                _make_big_query_json(crashes[5], False, True,
                                     _get_testcase_id(crashes[5])),
                '%s:bot:987:0' % crashes[5].key),
            big_query.Insert(
                _make_big_query_json(crashes[6], False, False, None),
                '%s:bot:987:1' % crashes[5].key)
        ]),
        mock.call(mock.ANY, [
            big_query.Insert(
                _make_big_query_json(crashes[7], False, True,
                                     _get_testcase_id(crashes[7])),
                '%s:bot:987:0' % crashes[7].key)
        ]),
    ])


class WriteCrashToBigQueryTest(unittest.TestCase):
  """Test write_crash_to_big_query."""

  def setUp(self):
    self.client = mock.Mock(spec_set=big_query.Client)
    helpers.patch(self, [
        'clusterfuzz._internal.system.environment.get_value',
        'clusterfuzz._internal.datastore.data_handler.get_project_name',
        'clusterfuzz._internal.google_cloud_utils.big_query.Client',
        'time.time',
    ])
    monitor.metrics_store().reset_for_testing()

    self.mock.get_project_name.return_value = 'some_project'
    self.mock.get_value.return_value = 'bot'
    self.mock.Client.return_value = self.client
    self.mock.time.return_value = 99
    self.crashes = [
        self._make_crash('c1'),
        self._make_crash('c2'),
        self._make_crash('c3')
    ]

    newly_created_testcase = mock.MagicMock()
    newly_created_testcase.key.id.return_value = 't'
    self.group = mock.MagicMock(
        crashes=self.crashes,
        main_crash=self.crashes[0],
        one_time_crasher_flag=False,
        newly_created_testcase=newly_created_testcase)
    self.group.is_new.return_value = True

  def _create_context(self, job_type, platform_id):
    return fuzz_task.Context(
        project_name='some_project',
        bot_name='bot',
        job_type=job_type,
        fuzz_target=data_types.FuzzTarget(engine='engine', binary='binary'),
        redzone=32,
        disable_ubsan=False,
        platform_id=platform_id,
        crash_revision=1234,
        fuzzer_name='engine',
        window_argument='windows_args',
        fuzzer_metadata={},
        testcases_metadata={},
        timeout_multiplier=1.0,
        test_timeout=5,
        thread_wait_timeout=6,
        data_directory='data')

  def _make_crash(self, state):
    crash = mock.Mock(
        crash_type='type',
        crash_state=state,
        crash_time=111,
        security_flag=True,
        key='key')
    return crash

  def _json(self, job, platform, state, new_flag, testcase_id):
    return {
        'crash_type': 'type',
        'crash_state': state,
        'created_at': 99,
        'platform': platform,
        'crash_time_in_ms': 111000,
        'parent_fuzzer_name': 'engine',
        'fuzzer_name': 'engine_binary',
        'job_type': job,
        'security_flag': True,
        'reproducible_flag': True,
        'revision': '1234',
        'new_flag': new_flag,
        'project': 'some_project',
        'testcase_id': testcase_id
    }

  def test_all_succeed(self):
    """Test writing succeeds."""
    self.client.insert.return_value = {}
    context = self._create_context('job', 'linux')
    fuzz_task.write_crashes_to_big_query(self.group, context)

    success_count = monitoring_metrics.BIG_QUERY_WRITE_COUNT.get({
        'success': True
    })
    failure_count = monitoring_metrics.BIG_QUERY_WRITE_COUNT.get({
        'success': False
    })

    self.assertEqual(3, success_count)
    self.assertEqual(0, failure_count)

    self.mock.Client.assert_called_once_with(
        dataset_id='main', table_id='crashes$19700101')
    self.client.insert.assert_called_once_with([
        big_query.Insert(
            self._json('job', 'linux', 'c1', True, 't'), 'key:bot:99:0'),
        big_query.Insert(
            self._json('job', 'linux', 'c2', False, None), 'key:bot:99:1'),
        big_query.Insert(
            self._json('job', 'linux', 'c3', False, None), 'key:bot:99:2')
    ])

  def test_succeed(self):
    """Test writing succeeds."""
    self.client.insert.return_value = {'insertErrors': [{'index': 1}]}
    context = self._create_context('job', 'linux')
    fuzz_task.write_crashes_to_big_query(self.group, context)

    success_count = monitoring_metrics.BIG_QUERY_WRITE_COUNT.get({
        'success': True
    })
    failure_count = monitoring_metrics.BIG_QUERY_WRITE_COUNT.get({
        'success': False
    })

    self.assertEqual(2, success_count)
    self.assertEqual(1, failure_count)

    self.mock.Client.assert_called_once_with(
        dataset_id='main', table_id='crashes$19700101')
    self.client.insert.assert_called_once_with([
        big_query.Insert(
            self._json('job', 'linux', 'c1', True, 't'), 'key:bot:99:0'),
        big_query.Insert(
            self._json('job', 'linux', 'c2', False, None), 'key:bot:99:1'),
        big_query.Insert(
            self._json('job', 'linux', 'c3', False, None), 'key:bot:99:2')
    ])

  def test_chromeos_platform(self):
    """Test ChromeOS platform is written in stats."""
    self.client.insert.return_value = {'insertErrors': [{'index': 1}]}
    context = self._create_context('job_chromeos', 'linux')
    fuzz_task.write_crashes_to_big_query(self.group, context)

    success_count = monitoring_metrics.BIG_QUERY_WRITE_COUNT.get({
        'success': True
    })
    failure_count = monitoring_metrics.BIG_QUERY_WRITE_COUNT.get({
        'success': False
    })

    self.assertEqual(2, success_count)
    self.assertEqual(1, failure_count)

    self.mock.Client.assert_called_once_with(
        dataset_id='main', table_id='crashes$19700101')
    self.client.insert.assert_called_once_with([
        big_query.Insert(
            self._json('job_chromeos', 'chrome', 'c1', True, 't'),
            'key:bot:99:0'),
        big_query.Insert(
            self._json('job_chromeos', 'chrome', 'c2', False, None),
            'key:bot:99:1'),
        big_query.Insert(
            self._json('job_chromeos', 'chrome', 'c3', False, None),
            'key:bot:99:2')
    ])

  def test_exception(self):
    """Test writing raising an exception."""
    self.client.insert.side_effect = Exception('error')
    context = self._create_context('job', 'linux')
    fuzz_task.write_crashes_to_big_query(self.group, context)

    success_count = monitoring_metrics.BIG_QUERY_WRITE_COUNT.get({
        'success': True
    })
    failure_count = monitoring_metrics.BIG_QUERY_WRITE_COUNT.get({
        'success': False
    })

    self.assertEqual(0, success_count)
    self.assertEqual(3, failure_count)


class ConvertGroupsToCrashesTest(object):
  """Test convert_groups_to_crashes."""

  def test_convert(self):
    """Test converting."""
    groups = [
        mock.Mock(
            crashes=[mock.Mock(), mock.Mock()],
            main_crash=mock.Mock(
                crash_type='t1', crash_state='s1', security_flag=True)),
        mock.Mock(
            crashes=[mock.Mock()],
            main_crash=mock.Mock(
                crash_type='t2', crash_state='s2', security_flag=False)),
    ]
    groups[0].is_new.return_value = False
    groups[1].is_new.return_value = True

    self.assertEqual([
        {
            'is_new': False,
            'count': 2,
            'crash_type': 't1',
            'crash_state': 's1',
            'security_flag': True
        },
        {
            'is_new': True,
            'count': 1,
            'crash_type': 't2',
            'crash_state': 's2',
            'security_flag': False
        },
    ], fuzz_task.convert_groups_to_crashes(groups))


class TestCorpusSync(fake_filesystem_unittest.TestCase):
  """Test corpus sync."""

  def setUp(self):
    """Setup for test corpus sync."""
    helpers.patch(self, [
        'clusterfuzz._internal.fuzzing.corpus_manager.FuzzTargetCorpus.rsync_to_disk',
        'clusterfuzz._internal.fuzzing.corpus_manager.FuzzTargetCorpus.upload_files',
        'clusterfuzz._internal.google_cloud_utils.storage.last_updated',
    ])

    helpers.patch_environ(self)

    os.environ['FAIL_RETRIES'] = '1'
    os.environ['CORPUS_BUCKET'] = 'bucket'

    self.mock.rsync_to_disk.return_value = True
    test_utils.set_up_pyfakefs(self)
    self.fs.create_dir('/dir')
    self.fs.create_dir('/dir1')

  def _write_corpus_files(self, *args, **kwargs):  # pylint: disable=unused-argument
    self.fs.create_file('/dir/a')
    self.fs.create_file('/dir/b')
    return True

  def test_sync(self):
    """Test corpus sync."""
    corpus = fuzz_task.GcsCorpus('parent', 'child', '/dir', '/dir1')

    self.mock.rsync_to_disk.side_effect = self._write_corpus_files
    self.assertTrue(corpus.sync_from_gcs())
    self.assertTrue(os.path.exists('/dir1/.child_sync'))
    self.assertEqual(('/dir',), self.mock.rsync_to_disk.call_args[0][1:])
    self.fs.create_file('/dir/c')
    self.assertListEqual(['/dir/c'], corpus.get_new_files())

    corpus.upload_files(corpus.get_new_files())
    self.assertEqual((['/dir/c'],), self.mock.upload_files.call_args[0][1:])

    self.assertListEqual([], corpus.get_new_files())

  def test_no_sync(self):
    """Test no corpus sync when bundle is not updated since last sync."""
    corpus = fuzz_task.GcsCorpus('parent', 'child', '/dir', '/dir1')

    utils.write_data_to_file(time.time(), '/dir1/.child_sync')
    self.mock.last_updated.return_value = (
        datetime.datetime.utcnow() - datetime.timedelta(days=1))
    self.assertTrue(corpus.sync_from_gcs())
    self.assertEqual(0, self.mock.rsync_to_disk.call_count)

  def test_sync_with_failed_last_update(self):
    """Test corpus sync when failed to get last update info from gcs."""
    corpus = fuzz_task.GcsCorpus('parent', 'child', '/dir', '/dir1')

    utils.write_data_to_file(time.time(), '/dir1/.child_sync')
    self.mock.last_updated.return_value = None
    self.assertTrue(corpus.sync_from_gcs())
    self.assertEqual(1, self.mock.rsync_to_disk.call_count)


@test_utils.with_cloud_emulators('datastore')
class DoBlackboxFuzzingTest(fake_filesystem_unittest.TestCase):
  """do_blackbox_fuzzing tests."""

  def setUp(self):
    """Setup for blackbox fuzzing test."""
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.random_element_from_list',
        'clusterfuzz._internal.base.utils.random_number',
        'clusterfuzz._internal.bot.fuzzers.engine_common.current_timestamp',
        'clusterfuzz._internal.bot.tasks.fuzz_task.pick_gestures',
        'clusterfuzz._internal.bot.testcase_manager.upload_log',
        'clusterfuzz._internal.bot.testcase_manager.upload_testcase',
        'clusterfuzz._internal.build_management.revisions.get_component_list',
        'clusterfuzz._internal.crash_analysis.crash_analyzer.is_crash',
        'clusterfuzz._internal.crash_analysis.stack_parsing.stack_analyzer.get_crash_data',
        'clusterfuzz._internal.datastore.ndb_init.context',
        'clusterfuzz._internal.metrics.fuzzer_stats.upload_stats',
        'random.random',
        'clusterfuzz._internal.system.process_handler.close_queue',
        'clusterfuzz._internal.system.process_handler.get_process',
        'clusterfuzz._internal.system.process_handler.get_queue',
        'clusterfuzz._internal.system.process_handler.run_process',
        'clusterfuzz._internal.system.process_handler.terminate_hung_threads',
        'clusterfuzz._internal.system.process_handler.'
        'terminate_stale_application_instances',
    ])

    os.environ['APP_ARGS'] = '-x'
    os.environ['APP_ARGS_APPEND_TESTCASE'] = 'True'
    os.environ['APP_DIR'] = '/app'
    os.environ['APP_NAME'] = 'app_1'
    os.environ['APP_PATH'] = '/app/app_1'
    os.environ['BOT_TMPDIR'] = '/tmp'
    os.environ['CRASH_STACKTRACES_DIR'] = '/crash'
    os.environ['ENABLE_GESTURES'] = 'False'
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['FUZZER_DIR'] = '/fuzzer'
    os.environ['INPUT_DIR'] = '/input'
    os.environ['JOB_NAME'] = 'asan_test'
    os.environ['MAX_FUZZ_THREADS'] = '1'
    os.environ['MAX_TESTCASES'] = '3'
    os.environ['RANDOM_SEED'] = '-r'
    os.environ['ROOT_DIR'] = '/root'
    os.environ['THREAD_ALIVE_CHECK_INTERVAL'] = '0.001'
    os.environ['THREAD_DELAY'] = '0.001'
    os.environ['USER_PROFILE_IN_MEMORY'] = 'True'

    test_utils.set_up_pyfakefs(self)
    self.fs.create_dir('/crash')
    self.fs.create_dir('/root/bot/logs')

    # Value picked as timeout multiplier.
    self.mock.random_element_from_list.return_value = 2.0
    # Choose window_arg, timeout multiplier, random seed.
    self.mock.random_number.side_effect = [0, 0, 3]
    # One trial profile for the session.
    self.mock.random.side_effect = [0.3, 0.3]
    self.mock.pick_gestures.return_value = []
    self.mock.get_component_list.return_value = [{
        'component': 'component',
        'link_text': 'rev',
    }]
    self.mock.current_timestamp.return_value = 0.0

    # Dummy output when running tests. E.g. exit code 0 and no output.
    self.mock.run_process.return_value = (0, 0, '')

    # Treat first and third run as crashed.
    self.mock.is_crash.side_effect = [True, False, True]

    self.mock.get_queue.return_value = queue.Queue()
    self.mock.get_process.return_value = threading.Thread

  def test_trials(self):
    """Test fuzzing session with trials."""
    data_types.Trial(app_name='app_1', probability=0.5, app_args='-y').put()
    data_types.Trial(app_name='app_1', probability=0.2, app_args='-z').put()

    session = fuzz_task.FuzzingSession('fantasy_fuzz', 'asan_test', 10)
    self.assertEqual(20, session.test_timeout)

    # Mock out actual test-case generation for 3 tests.
    session.generate_blackbox_testcases = mock.MagicMock()
    expected_testcase_file_paths = ['/tests/0', '/tests/1', '/tests/2']
    session.generate_blackbox_testcases.return_value = (
        False, expected_testcase_file_paths, None, {
            'fuzzer_binary_name': 'fantasy_fuzz'
        })

    fuzzer = data_types.Fuzzer()
    fuzzer.name = 'fantasy_fuzz'

    fuzzer_metadata, testcase_file_paths, testcases_metadata, crashes = (
        session.do_blackbox_fuzzing(fuzzer, '/fake-fuzz-dir', 'asan_test'))

    self.assertEqual({'fuzzer_binary_name': 'fantasy_fuzz'}, fuzzer_metadata)
    self.assertEqual(expected_testcase_file_paths, testcase_file_paths)
    self.assertEqual(
        dict((t, {
            'gestures': []
        }) for t in expected_testcase_file_paths), testcases_metadata)

    self.assertEqual(3, len(self.mock.is_crash.call_args_list))

    # Verify the three test runs are called with the correct arguments.
    calls = self.mock.run_process.call_args_list
    self.assertEqual(3, len(calls))
    self.assertEqual('/app/app_1 -r=3 -x -y /tests/0', calls[0][0][0])
    self.assertEqual('/app/app_1 -r=3 -x -y /tests/1', calls[1][0][0])
    self.assertEqual('/app/app_1 -r=3 -x -y /tests/2', calls[2][0][0])

    # Verify the two crashes store the correct arguments.
    self.assertEqual(2, len(crashes))
    self.assertEqual('/app/app_1 -r=3 -x -y /tests/0',
                     crashes[0].application_command_line)
    self.assertEqual('/app/app_1 -r=3 -x -y /tests/2',
                     crashes[1].application_command_line)


@test_utils.with_cloud_emulators('datastore')
class DoEngineFuzzingTest(fake_filesystem_unittest.TestCase):
  """do_engine_fuzzing tests."""

  def setUp(self):
    """Setup for do engine fuzzing test."""
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.engine_common.current_timestamp',
        'clusterfuzz._internal.bot.tasks.fuzz_task.GcsCorpus.sync_from_gcs',
        'clusterfuzz._internal.bot.tasks.fuzz_task.GcsCorpus.upload_files',
        'clusterfuzz._internal.build_management.revisions.get_component_list',
        'clusterfuzz._internal.bot.testcase_manager.upload_log',
        'clusterfuzz._internal.bot.testcase_manager.upload_testcase',
        'clusterfuzz._internal.metrics.fuzzer_stats.upload_stats',
    ])
    test_utils.set_up_pyfakefs(self)

    os.environ['JOB_NAME'] = 'libfuzzer_asan_test'
    os.environ['FUZZ_INPUTS'] = '/fuzz-inputs'
    os.environ['FUZZ_INPUTS_DISK'] = '/fuzz-inputs-disk'
    os.environ['BUILD_DIR'] = '/build_dir'
    os.environ['MAX_TESTCASES'] = '2'
    os.environ['AUTOMATIC_LABELS'] = 'auto_label,auto_label1'
    os.environ['AUTOMATIC_COMPONENTS'] = 'auto_component,auto_component1'

    self.fs.create_file('/build_dir/test_target')
    self.fs.create_file(
        '/build_dir/test_target.labels', contents='label1\nlabel2')
    self.fs.create_file(
        '/build_dir/test_target.owners', contents='owner1@email.com')
    self.fs.create_file(
        '/build_dir/test_target.components', contents='component1\ncomponent2')
    self.fs.create_file('/input')

    self.mock.sync_from_gcs.return_value = True
    self.mock.upload_files.return_value = True
    self.mock.get_component_list.return_value = [{
        'component': 'component',
        'link_text': 'rev',
    }]
    self.mock.current_timestamp.return_value = 0.0

  def test_basic(self):
    """Test basic fuzzing session."""
    session = fuzz_task.FuzzingSession('libFuzzer', 'libfuzzer_asan_test', 60)
    session.testcase_directory = os.environ['FUZZ_INPUTS']
    session.data_directory = '/data_dir'

    os.environ['FUZZ_TARGET'] = 'test_target'
    os.environ['APP_REVISION'] = '1'
    os.environ['FUZZ_TEST_TIMEOUT'] = '2000'
    os.environ['BOT_NAME'] = 'hostname.company.com'

    expected_crashes = [engine.Crash('/input', 'stack', ['args'], 1.0)]

    engine_impl = mock.Mock()
    engine_impl.name = 'libFuzzer'
    engine_impl.prepare.return_value = engine.FuzzOptions(
        '/corpus', ['arg'], {
            'strategy_1': 1,
            'strategy_2': 50,
        })
    engine_impl.fuzz.side_effect = lambda *_: engine.FuzzResult(
        'logs', ['cmd'], expected_crashes, {'stat': 1}, 42.0)
    engine_impl.fuzz_additional_processing_timeout.return_value = 1337

    crashes, fuzzer_metadata = session.do_engine_fuzzing(engine_impl)

    engine_impl.fuzz.assert_called_with('/build_dir/test_target',
                                        engine_impl.prepare.return_value,
                                        '/fuzz-inputs', 663)
    self.assertDictEqual({
        'fuzzer_binary_name':
            'test_target',
        'issue_components':
            'component1,component2,auto_component,auto_component1',
        'issue_labels':
            'label1,label2,auto_label,auto_label1',
        'issue_owners':
            'owner1@email.com',
    }, fuzzer_metadata)

    log_time = datetime.datetime(1970, 1, 1, 0, 0)
    log_call = mock.call(
        b'Component revisions (build r1):\n'
        b'component: rev\n\nBot name: hostname.company.com\n'
        b'Return code: 1\n\n'
        b'Command: cmd\nTime ran: 42.0\n\n'
        b'logs\n'
        b'cf::fuzzing_strategies: strategy_1:1,strategy_2:50', log_time)
    self.mock.upload_log.assert_has_calls([log_call, log_call])
    self.mock.upload_testcase.assert_has_calls([
        mock.call('/input', log_time),
        mock.call('/input', log_time),
    ])

    self.assertEqual(2, len(crashes))
    for i in range(2):
      self.assertEqual('/input', crashes[i].file_path)
      self.assertEqual(1, crashes[i].return_code)
      self.assertEqual('stack', crashes[i].unsymbolized_crash_stacktrace)
      self.assertEqual(1.0, crashes[i].crash_time)
      self.assertEqual('args', crashes[i].arguments)

    for i in range(2):
      upload_args = self.mock.upload_stats.call_args_list[i][0][0]
      testcase_run = upload_args[0]
      self.assertDictEqual({
          'build_revision': 1,
          'command': ['cmd'],
          'fuzzer': u'libFuzzer_test_target',
          'job': 'libfuzzer_asan_test',
          'kind': 'TestcaseRun',
          'stat': 1,
          'strategy_strategy_1': 1,
          'strategy_strategy_2': 50,
          'timestamp': 0.0,
      }, testcase_run.data)


class UntrustedRunEngineFuzzerTest(
    untrusted_runner_helpers.UntrustedRunnerIntegrationTest):
  """Engine fuzzing tests for untrusted."""

  def setUp(self):
    """Set up."""
    super().setUp()
    environment.set_value('JOB_NAME', 'libfuzzer_asan_job')

    job = data_types.Job(
        name='libfuzzer_asan_job',
        environment_string=(
            'RELEASE_BUILD_BUCKET_PATH = '
            'gs://clusterfuzz-test-data/test_libfuzzer_builds/'
            'test-libfuzzer-build-([0-9]+).zip\n'
            'REVISION_VARS_URL = https://commondatastorage.googleapis.com/'
            'clusterfuzz-test-data/test_libfuzzer_builds/'
            'test-libfuzzer-build-%s.srcmap.json\n'))
    job.put()

    self.temp_dir = tempfile.mkdtemp(dir=environment.get_value('FUZZ_INPUTS'))
    environment.set_value('USE_MINIJAIL', False)

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.temp_dir, ignore_errors=True)

  def test_run_engine_fuzzer(self):
    """Test running engine fuzzer."""
    self._setup_env(job_type='libfuzzer_asan_job')
    environment.set_value('FUZZ_TEST_TIMEOUT', 3600)
    environment.set_value('STRATEGY_SELECTION_METHOD', 'multi_armed_bandit')
    environment.set_value(
        'STRATEGY_SELECTION_DISTRIBUTION',
        '[{"strategy_name": "value_profile", '
        '"probability": 1.0, "engine": "libFuzzer"}]')

    build_manager.setup_build()
    corpus_directory = os.path.join(self.temp_dir, 'corpus')
    testcase_directory = os.path.join(self.temp_dir, 'artifacts')
    os.makedirs(file_host.rebase_to_worker_root(corpus_directory))
    os.makedirs(file_host.rebase_to_worker_root(testcase_directory))

    result, fuzzer_metadata, strategies = fuzz_task.run_engine_fuzzer(
        libfuzzer_engine.LibFuzzerEngine(), 'test_fuzzer', corpus_directory,
        testcase_directory)
    self.assertIn(
        'ERROR: AddressSanitizer: SEGV on unknown address 0x000000000000',
        result.logs)
    self.assertEqual(1, len(result.crashes))
    self.assertTrue(result.crashes[0].input_path.startswith(
        os.environ['ROOT_DIR']))
    self.assertTrue(os.path.exists(result.crashes[0].input_path))
    self.assertIsInstance(
        result.stats.get('number_of_executed_units'), six.integer_types)
    self.assertIsInstance(result.stats.get('oom_count'), six.integer_types)
    self.assertIsInstance(
        result.stats.get('strategy_selection_method'), six.string_types)

    self.assertDictEqual({'fuzzer_binary_name': 'test_fuzzer'}, fuzzer_metadata)
    self.assertDictEqual({'value_profile': 1}, strategies)


class AddIssueMetadataFromEnvironmentTest(unittest.TestCase):
  """Tests for _add_issue_metadata_from_environment."""

  def setUp(self):
    helpers.patch_environ(self)

  def test_add_no_existing(self):
    """Test adding issue metadata when there are none existing."""
    os.environ['AUTOMATIC_LABELS'] = 'auto_label'
    os.environ['AUTOMATIC_LABELS_1'] = 'auto_label1'
    os.environ['AUTOMATIC_COMPONENTS'] = 'auto_component'
    os.environ['AUTOMATIC_COMPONENTS_1'] = 'auto_component1'

    metadata = {}
    fuzz_task._add_issue_metadata_from_environment(metadata)
    self.assertDictEqual({
        'issue_components': 'auto_component,auto_component1',
        'issue_labels': 'auto_label,auto_label1',
    }, metadata)

  def test_add_append(self):
    """Test adding issue metadata when there are already existing metadata."""
    os.environ['AUTOMATIC_LABELS'] = 'auto_label'
    os.environ['AUTOMATIC_LABELS_1'] = 'auto_label1'
    os.environ['AUTOMATIC_COMPONENTS'] = 'auto_component'
    os.environ['AUTOMATIC_COMPONENTS_1'] = 'auto_component1'

    metadata = {
        'issue_components': 'existing_component',
        'issue_labels': 'existing_label'
    }
    fuzz_task._add_issue_metadata_from_environment(metadata)
    self.assertDictEqual({
        'issue_components':
            'existing_component,auto_component,auto_component1',
        'issue_labels':
            'existing_label,auto_label,auto_label1',
    }, metadata)

  def test_add_numeric(self):
    """Tests adding a numeric label."""
    os.environ['AUTOMATIC_LABELS'] = '123'

    metadata = {}
    fuzz_task._add_issue_metadata_from_environment(metadata)
    self.assertDictEqual({
        'issue_labels': '123',
    }, metadata)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Memoize caches the result of methods."""

import collections
import functools
import json
import threading

import six

from clusterfuzz._internal.base import persistent_cache
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system.environment import appengine_noop
from clusterfuzz._internal.system.environment import bot_noop
from clusterfuzz._internal.system.environment import local_noop

# Thead local globals.
_local = threading.local()

_DEFAULT_REDIS_HOST = 'localhost'
_DEFAULT_REDIS_PORT = 6379


def _redis_client():
  """Get the redis client."""
  import redis

  if hasattr(_local, 'redis'):
    return _local.redis

  host = environment.get_value('REDIS_HOST', _DEFAULT_REDIS_HOST)
  port = environment.get_value('REDIS_PORT', _DEFAULT_REDIS_PORT)

  _local.redis = redis.Redis(host=host, port=port)
  return _local.redis


class FifoInMemory(object):
  """In-memory caching engine."""

  def __init__(self, capacity):
    self.capacity = capacity
    self.lock = threading.Lock()
    self._cache = None

  @property
  def cache(self):
    """Get the cache backing. None may be returned."""
    if self._cache is None:
      self._cache = collections.OrderedDict()

    return self._cache

  def put(self, key, value):
    """Put (key, value) into cache."""
    if self.cache is None:
      return

    # Lock to avoid race condition in popitem.
    self.lock.acquire()

    if len(self.cache) >= self.capacity:
      self.cache.popitem(last=False)

    self.cache[key] = value

    self.lock.release()

  def get(self, key):
    """Get the value from cache."""
    if self.cache is None:
      return None

    return self.cache.get(key)

  def get_key(self, func, args, kwargs):
    """Get a key name based on function, arguments and keyword arguments."""
    return _default_key(func, args, kwargs)


class FifoOnDisk(object):
  """On-disk caching engine."""

  def __init__(self, capacity):
    self.capacity = capacity
    self.keys = []
    self.lock = threading.Lock()

  @appengine_noop
  def put(self, key, value):
    """Put (key, value) into cache."""
    # Lock to avoid race condition in pop.
    self.lock.acquire()

    if len(self.keys) >= self.capacity:
      key_to_remove = self.keys.pop(0)
      persistent_cache.delete_value(key_to_remove)

    persistent_cache.set_value(key, value)
    self.keys.append(key)

    self.lock.release()

  @appengine_noop
  def get(self, key):
    """Get the value from cache."""
    return persistent_cache.get_value(key)

  def get_key(self, func, args, kwargs):
    """Get a key name based on function, arguments and keyword arguments."""
    return _default_key(func, args, kwargs)


class Memcache(object):
  """Memcache caching engine."""

  def __init__(self, ttl_in_seconds, key_fn=None):
    self.ttl_in_seconds = ttl_in_seconds
    self.key_fn = key_fn or _default_key

  @local_noop
  @bot_noop
  def put(self, key, value):
    """Put (key, value) into cache."""
    import redis
    try:
      _redis_client().set(
          json.dumps(key), json.dumps(value), ex=self.ttl_in_seconds)
    except redis.RedisError:
      logs.log_error('Failed to store key in cache.', key=key, value=value)

  @local_noop
  @bot_noop
  def get(self, key):
    """Get the value from cache."""
    import redis
    try:
      value_raw = _redis_client().get(json.dumps(key))
    except redis.RedisError:
      logs.log_error('Failed to retrieve key from cache.', key=key)
      return None

    if value_raw is None:
      return value_raw

    return json.loads(value_raw)

  def get_key(self, func, args, kwargs):
    return self.key_fn(func, args, kwargs)


def _default_key(func, args, kwargs):
  """Get a key name based on function, arguments and keyword arguments."""
  # Use unicode instead of str where possible. This makes it less likely to
  # have false misses.
  args = tuple(arg if not isinstance(arg, str) else str(arg) for arg in args)

  kwargs = {
      key: value if not isinstance(value, str) else str(value)
      for key, value in six.iteritems(kwargs)
  }

  return 'memoize:%s' % [func.__name__, args, sorted(kwargs.items())]


def wrap(engine):
  """Decorator for caching the result of method calls. Arguments must
    be hashable. None is not cached because we don't tell the difference
    between having None and not having a key."""

  def decorator(func):
    """Decorator function."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
      """Wrapper function."""
      force_update = kwargs.pop('__memoize_force__', False)

      key = engine.get_key(func, args, kwargs)
      result = engine.get(key)

      if result is not None and not force_update:
        return result

      result = func(*args, **kwargs)
      engine.put(key, result)
      return result

    return wrapper

  return decorator



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for syzkaller engine."""
# pylint: disable=unused-argument

import os
import shutil
import unittest

from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils

TEST_PATH = os.path.abspath(os.path.dirname(__file__))
DATA_DIR = os.path.join(TEST_PATH, 'test_data')
TEMP_DIR = os.path.join(TEST_PATH, 'temp')


def clear_temp_dir():
  """Clear temp directory."""
  if os.path.exists(TEMP_DIR):
    shutil.rmtree(TEMP_DIR)

  os.mkdir(TEMP_DIR)


@test_utils.integration
class IntegrationTest(unittest.TestCase):
  """Integration tests."""

  def setUp(self):
    test_helpers.patch_environ(self)

    os.environ['BUILD_DIR'] = DATA_DIR



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for redirecting to the issue url (given a testcase). See
  crbug.com/665652 on why we need it."""

from handlers import base_handler
from libs import helpers
from libs.issue_management import issue_tracker_utils


class Handler(base_handler.Handler):
  """Handler that redirects user to the issue URL."""

  def get(self, testcase_id=None):
    """Redirect user to the correct URL."""
    testcase = helpers.get_testcase(testcase_id)
    issue_url = helpers.get_or_exit(
        lambda: issue_tracker_utils.get_issue_url(testcase),
        'Issue tracker for testcase (id=%s) is not found.' % testcase_id,
        'Failed to get the issue tracker URL.')

    return self.redirect(issue_url)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Report upload task."""

import time

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.chrome import crash_uploader
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment


def execute_task(*_):
  """Execute the report uploads."""
  logs.log('Uploading pending reports.')

  # Get metadata for reports requiring upload.
  reports_metadata = ndb_utils.get_all_from_query(
      data_types.ReportMetadata.query(
          ndb_utils.is_false(data_types.ReportMetadata.is_uploaded)))
  reports_metadata = list(reports_metadata)
  if not reports_metadata:
    logs.log('No reports that need upload found.')
    return

  environment.set_value('UPLOAD_MODE', 'prod')

  # Otherwise, upload corresponding reports.
  logs.log('Uploading reports for testcases: %s' % str(
      [report.testcase_id for report in reports_metadata]))

  report_metadata_to_delete = []
  for report_metadata in reports_metadata:
    # Convert metadata back into actual report.
    crash_info = crash_uploader.crash_report_info_from_metadata(report_metadata)
    testcase_id = report_metadata.testcase_id

    try:
      _ = data_handler.get_testcase_by_id(testcase_id)
    except errors.InvalidTestcaseError:
      logs.log_warn('Could not find testcase %s.' % testcase_id)
      report_metadata_to_delete.append(report_metadata.key)
      continue

    # Upload the report and update the corresponding testcase info.
    logs.log('Processing testcase %s for crash upload.' % testcase_id)
    crash_report_id = crash_info.upload()
    if crash_report_id is None:
      logs.log_error(
          'Crash upload for testcase %s failed, retry later.' % testcase_id)
      continue

    # Update the report metadata to indicate successful upload.
    report_metadata.crash_report_id = crash_report_id
    report_metadata.is_uploaded = True
    report_metadata.put()

    logs.log('Uploaded testcase %s to crash, got back report id %s.' %
             (testcase_id, crash_report_id))
    time.sleep(1)

  # Delete report metadata entries where testcase does not exist anymore or
  # upload is not supported.
  if report_metadata_to_delete:
    ndb_utils.delete_multi(report_metadata_to_delete)

  # Log done with uploads.
  # Deletion happens in batches in cleanup_task, so that in case of error there
  # is some buffer for looking at stored ReportMetadata in the meantime.
  logs.log('Finished uploading crash reports.')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for module management."""

# Do not add any imports to non-standard modules here.
import os
import site
import sys


def _config_modules_directory(root_directory):
  """Get the config modules directory."""
  config_dir = os.getenv('CONFIG_DIR_OVERRIDE')
  if not config_dir:
    config_dir = os.path.join(root_directory, 'src', 'appengine', 'config')

  return os.path.join(config_dir, 'modules')


def _patch_appengine_modules_for_bots():
  """Patch out App Engine reliant behaviour from bots."""
  if os.getenv('SERVER_SOFTWARE'):
    # Not applicable on App Engine.
    return

  # google.auth uses App Engine credentials based on importability of
  # google.appengine.api.app_identity.
  try:
    from google.auth import app_engine as auth_app_engine
    if auth_app_engine.app_identity:
      auth_app_engine.app_identity = None
  except ImportError:
    pass


def fix_module_search_paths():
  """Add directories that we must be able to import from to path."""
  root_directory = os.environ['ROOT_DIR']
  source_directory = os.path.join(root_directory, 'src')

  python_path = os.getenv('PYTHONPATH', '').split(os.pathsep)

  third_party_libraries_directory = os.path.join(source_directory,
                                                 'third_party')
  config_modules_directory = _config_modules_directory(root_directory)

  if (os.path.exists(config_modules_directory) and
      config_modules_directory not in sys.path):
    sys.path.insert(0, config_modules_directory)
    python_path.insert(0, config_modules_directory)

  if third_party_libraries_directory not in sys.path:
    sys.path.insert(0, third_party_libraries_directory)
    python_path.insert(0, third_party_libraries_directory)

  if source_directory not in sys.path:
    sys.path.insert(0, source_directory)
    python_path.insert(0, source_directory)

  os.environ['PYTHONPATH'] = os.pathsep.join(python_path)

  # Add site directory to make from imports work in google namespace.
  site.addsitedir(third_party_libraries_directory)

  # TODO(ochang): Remove this once SDK is removed from images.
  _patch_appengine_modules_for_bots()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for ML RNN generator."""

import os
import sys
import tempfile
import unittest

from clusterfuzz._internal.bot.fuzzers.ml.rnn import constants
from clusterfuzz._internal.bot.fuzzers.ml.rnn import generator
from clusterfuzz._internal.system import shell
from clusterfuzz._internal.tests.test_libs import test_utils

DATA_DIRECTORY = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'data'))

# Parameters for the demo model.
MODEL_NAME = 'rnn'
MODEL_LAYER_SIZE = 1
MODEL_STATE_SIZE = 2


@unittest.skipIf(not os.getenv('ML_TESTS'), 'ML_TESTS=1 must be set')
@test_utils.integration
class MLRnnGeneratorIntegrationTest(unittest.TestCase):
  """ML RNN generator tests."""

  def setUp(self):
    self.model_directory = os.path.join(DATA_DIRECTORY, 'model')
    self.model_path = os.path.join(self.model_directory, MODEL_NAME)

    self.input_directory = os.path.join(DATA_DIRECTORY, 'input')
    self.output_directory = tempfile.mkdtemp()

    self.empty_directory = tempfile.mkdtemp()

  def tearDown(self):
    shell.remove_directory(self.output_directory)
    shell.remove_directory(self.empty_directory)

  def test_generate(self):
    """Test generate specified number of inputs."""
    # Set a large timeout value and a small count value to avoid timeout.
    timeout = 20
    expected_count = 2

    result = generator.run(
        self.input_directory,
        self.output_directory,
        self.model_path,
        timeout,
        generation_count=expected_count,
        hidden_state_size=MODEL_STATE_SIZE,
        hidden_layer_size=MODEL_LAYER_SIZE)

    # Process exits normally and no timeout.
    self.assertEqual(result.return_code, constants.ExitCode.SUCCESS)
    self.assertFalse(result.timed_out)

    actual_count = shell.get_directory_file_count(self.output_directory)
    self.assertEqual(expected_count, actual_count)

  def test_empty_corpus(self):
    """Test generation should abort for empty corpus."""
    # Set a large timeout value and a small count value to avoid timeout.
    timeout = 20
    expected_count = 2

    result = generator.run(
        self.empty_directory,
        self.output_directory,
        self.model_path,
        timeout,
        generation_count=expected_count,
        hidden_state_size=MODEL_STATE_SIZE,
        hidden_layer_size=MODEL_LAYER_SIZE)

    self.assertEqual(result.return_code, constants.ExitCode.CORPUS_TOO_SMALL)
    self.assertFalse(result.timed_out)

    # No new units.
    actual_count = shell.get_directory_file_count(self.output_directory)
    self.assertEqual(actual_count, 0)

  def test_invalid_model(self):
    """Test TensorFlow should throw exception if model does not match."""
    # Set a large timeout value and a small count value to avoid timeout.
    timeout = 20
    expected_count = 2

    # Change model parameters to make demo model invalid.
    invalid_state_size = 8

    result = generator.run(
        self.input_directory,
        self.output_directory,
        self.model_path,
        timeout,
        generation_count=expected_count,
        hidden_state_size=invalid_state_size,
        hidden_layer_size=MODEL_LAYER_SIZE)

    self.assertEqual(result.return_code, constants.ExitCode.TENSORFLOW_ERROR)
    self.assertFalse(result.timed_out)

    # No new units.
    actual_count = shell.get_directory_file_count(self.output_directory)
    self.assertEqual(actual_count, 0)

  @test_utils.slow
  def test_timeout(self):
    """Test timeout case in generation."""
    # Set a small timeout value and a large count value to trigger timeout.
    # Note that timeout cannot be set too small since it takes time to
    # start generator. If this test failed please increase timeout value.
    timeout = 10

    result = generator.run(
        self.input_directory,
        self.output_directory,
        self.model_path,
        timeout,
        generation_count=sys.maxsize,
        hidden_state_size=MODEL_STATE_SIZE,
        hidden_layer_size=MODEL_LAYER_SIZE)

    # Process timed out.
    self.assertNotEqual(result.return_code, constants.ExitCode.SUCCESS)
    self.assertTrue(result.timed_out)

    actual_count = shell.get_directory_file_count(self.output_directory)
    self.assertGreater(actual_count, 0)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions for running commands on Fuchsia devices."""

# TODO(flowerhack): Re-enable this check once functions below are implemented.
# pylint: disable=unused-argument

import os
import shlex
import shutil
import socket
import subprocess
import tempfile
import time

from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.platforms.fuchsia import errors
from clusterfuzz._internal.platforms.fuchsia.util.device import Device
from clusterfuzz._internal.platforms.fuchsia.util.fuzzer import Fuzzer
from clusterfuzz._internal.platforms.fuchsia.util.host import Host
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.system import process_handler

_QEMU_WAIT_SECONDS = 60


def _fetch_qemu_vars():
  """
  Returns a dictionary with variables necessary for configuring and running
  Fuchsia via QEMU.

  This function assumes the following layout for
  fuchsia_resources_dir:

  * /qemu-for-fuchsia/*
  * /.ssh/*
  * target/x64/fvm.blk
  * target/x64/fuchsia.zbi
  * target/x64/multiboot.bin

  * build/out/default/fuzzers.json
  * build/out/default/ids.txt
  * build/out/default.zircon/tools/*
  * build/zircon/prebuilt/downloads/symbolize
  * build/buildtools/linux-x64/clang/bin/llvm-symbolizer

  If these locations change in the build, they should be changed here as well.
  """
  qemu_vars = {}

  # First, ensure we have a resources directory.
  fuchsia_resources_dir = environment.get_value('FUCHSIA_RESOURCES_DIR')
  if not fuchsia_resources_dir:
    raise errors.FuchsiaConfigError('Could not find FUCHSIA_RESOURCES_DIR')

  # Then, save and chmod the associated paths.
  qemu_vars['fuchsia_resources_dir'] = fuchsia_resources_dir
  qemu_vars['qemu_path'] = os.path.join(
      fuchsia_resources_dir, 'qemu-for-fuchsia', 'bin', 'qemu-system-x86_64')
  os.chmod(qemu_vars['qemu_path'], 0o550)
  qemu_vars['kernel_path'] = os.path.join(fuchsia_resources_dir, 'target',
                                          'x64', 'multiboot.bin')
  os.chmod(qemu_vars['kernel_path'], 0o644)
  qemu_vars['pkey_path'] = os.path.join(fuchsia_resources_dir, '.ssh', 'pkey')
  os.chmod(qemu_vars['pkey_path'], 0o400)
  qemu_vars['sharefiles_path'] = os.path.join(
      fuchsia_resources_dir, 'qemu-for-fuchsia', 'share', 'qemu')
  qemu_vars['drive_path'] = os.path.join(fuchsia_resources_dir, 'target', 'x64',
                                         'fvm-extended.blk')
  qemu_vars['orig_drive_path'] = os.path.join(fuchsia_resources_dir, 'target',
                                              'x64', 'fvm.blk')
  os.chmod(qemu_vars['orig_drive_path'], 0o644)
  qemu_vars['fuchsia_zbi'] = os.path.join(fuchsia_resources_dir, 'target',
                                          'x64', 'fuchsia.zbi')
  qemu_vars['initrd_path'] = os.path.join(fuchsia_resources_dir,
                                          'fuchsia-ssh.zbi')
  return qemu_vars


def initial_qemu_setup():
  """Performs one-time setup necessary to subsequently run Fuchsia QEMU VMs.
  This only needs to be called once per build setup, and will do nothing if
  called multiple times.
  This function does not run a VM, merely performs setup for a VM.
  """
  qemu_vars = _fetch_qemu_vars()

  # Exit early if it appears we've already been called for this build
  if os.path.exists(qemu_vars['initrd_path']):
    return

  extend_fvm(qemu_vars['fuchsia_resources_dir'], qemu_vars['orig_drive_path'],
             qemu_vars['drive_path'])
  add_keys_to_zbi(qemu_vars['fuchsia_resources_dir'], qemu_vars['initrd_path'],
                  qemu_vars['fuchsia_zbi'])


class QemuError(Exception):
  """Error for errors handling QEMU."""


class QemuProcess(object):
  """A QemuProcess encapsulates the creation, running, and destruction
  of Fuchsia QEMU processes."""

  # For now, use a system-global log path so we don't need to pass a tempfile
  # path around everywhere. We use a class constant so it can be accessed
  # without an instance, but defer initialization until the constructor so that
  # it isn't run in non-Fuchsia contexts.
  LOG_PATH = None

  def __init__(self):
    self.process_runner = None
    self.popen = None
    self.logfile = None

    if not QemuProcess.LOG_PATH:
      QemuProcess.LOG_PATH = os.path.join(tempfile.gettempdir(),
                                          'fuchsia-qemu-log')

  def create(self):
    """Configures a QEMU process which can subsequently be `run`.

    Assumes that initial_qemu_setup was already called exactly once.
    """
    qemu_vars = _fetch_qemu_vars()

    # Get a free port for the VM, so we can SSH in later.
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(('localhost', 0))
    _, port = tcp.getsockname()
    tcp.close()
    # Fuzzing jobs that SSH into the QEMU VM need access to this env var.
    environment.set_value('FUCHSIA_PORTNUM', port)
    environment.set_value('FUCHSIA_RESOURCES_DIR',
                          qemu_vars['fuchsia_resources_dir'])

    # yapf: disable
    qemu_args = [
        '-m', '3072',
        '-nographic',
        '-kernel', qemu_vars['kernel_path'],
        '-initrd', qemu_vars['initrd_path'],
        '-smp', '4',
        '-drive',
        ('file=' + qemu_vars['drive_path'] + ',format=raw,if=none,'
         'id=blobstore'),
        '-device', 'virtio-blk-pci,drive=blobstore',
        '-monitor', 'none',
        '-append', 'kernel.serial=legacy TERM=dumb',
        '-machine', 'q35',
        '-display', 'none',
        '-netdev',
        ('user,id=net0,net=192.168.3.0/24,dhcpstart=192.168.3.9,'
         'host=192.168.3.2,hostfwd=tcp::') + str(port) + '-:22',
        '-device', 'e1000,netdev=net0,mac=52:54:00:63:5e:7b',
        '-L', qemu_vars['sharefiles_path']
    ]
    # yapf: enable

    # Detecting KVM is tricky, so use an environment variable to determine
    # whether to turn it on or not.
    if environment.get_value('FUCHSIA_USE_KVM'):
      # In builds before fxrev.dev/375343, a bug prevents booting with newer
      # versions of KVM. On some of these older builds,
      # `kernel.x86.disable-spec-mitigations` also doesn't work as
      # expected, so we work around this by selecting a CPU type where the
      # speculation mitigation will not applied.
      if environment.get_value('APP_REVISION') < 20200414210423:
        qemu_args.extend(['-cpu', 'Opteron_G5,+invtsc'])
      else:
        qemu_args.extend(['-cpu', 'host,migratable=no,+invtsc'])
      qemu_args.append('-enable-kvm')
    else:
      # Can't use host CPU since we don't necessarily have KVM on the machine.
      # Emulate a Haswell CPU with a few feature toggles. This mirrors the most
      # common configuration for Fuchsia VMs when using in-tree tools.
      qemu_args.extend(['-cpu', 'Haswell,+smap,-check,-fsgsbase'])

    # Get the list of fuzzers for ClusterFuzz to choose from.
    host = Host.from_dir(
        os.path.join(qemu_vars['fuchsia_resources_dir'], 'build', 'out',
                     'default'))
    Device(host, 'localhost', str(port))
    Fuzzer.filter(host.fuzzers, '')

    # Fuzzing jobs that SSH into the QEMU VM need access to this env var.
    environment.set_value('FUCHSIA_PKEY_PATH', qemu_vars['pkey_path'])
    logs.log('Ready to run QEMU. Command: ' + qemu_vars['qemu_path'] + ' ' +
             ' '.join(shlex.quote(arg) for arg in qemu_args))
    self.process_runner = new_process.ProcessRunner(qemu_vars['qemu_path'],
                                                    qemu_args)

  def run(self):
    """Actually runs a QEMU VM, assuming `create` has already been called."""
    if not self.process_runner:
      raise QemuError('Attempted to `run` QEMU VM before calling `create`')

    # pylint: disable=consider-using-with
    self.logfile = open(QemuProcess.LOG_PATH, 'wb')
    self.popen = self.process_runner.run(
        stdout=self.logfile, stderr=subprocess.PIPE)
    time.sleep(_QEMU_WAIT_SECONDS)

  def kill(self):
    """ Kills the currently-running QEMU VM, if there is one. """
    if self.popen:
      self.popen.kill()
      self.popen = None

    if self.logfile:
      self.logfile.close()
      self.logfile = None


def start_qemu():
  """Start qemu."""
  qemu = QemuProcess()
  qemu.create()
  qemu.run()


def stop_qemu():
  """Stop qemu."""
  process_handler.terminate_processes_matching_names('qemu-system-x86_64')


def extend_fvm(fuchsia_resources_dir, orig_drive_path, drive_path):
  """The FVM is minimally sized to begin with; make an extended copy
  of it to make room for ephemeral packages etc."""
  fvm_tool_path = os.path.join(fuchsia_resources_dir, 'build', 'out',
                               'default.zircon', 'tools', 'fvm')
  os.chmod(fvm_tool_path, 0o500)

  # Since the fvm tool modifies the image in place, make a copy so the build
  # isn't mutated (required for running undercoat on a cached build previously
  # affected by this legacy codepath)
  shutil.copy(orig_drive_path, drive_path)

  process = new_process.ProcessRunner(fvm_tool_path,
                                      [drive_path, 'extend', '--length', '3G'])
  result = process.run_and_wait()
  if result.return_code or result.timed_out:
    raise errors.FuchsiaSdkError('Failed to extend FVM: ' + result.output)

  os.chmod(drive_path, 0o644)


def add_keys_to_zbi(fuchsia_resources_dir, initrd_path, fuchsia_zbi):
  """Adds keys to the ZBI so we can SSH into it. See:
  fuchsia.googlesource.com/fuchsia/+/refs/heads/master/sdk/docs/ssh.md"""
  zbi_tool = os.path.join(fuchsia_resources_dir, 'build', 'out',
                          'default.zircon', 'tools', 'zbi')
  os.chmod(zbi_tool, 0o500)
  authorized_keys_path = os.path.join(fuchsia_resources_dir, '.ssh',
                                      'authorized_keys')
  process = new_process.ProcessRunner(zbi_tool, [
      '-o', initrd_path, fuchsia_zbi, '-e',
      'data/ssh/authorized_keys=' + authorized_keys_path
  ])
  result = process.run_and_wait()
  if result.return_code or result.timed_out:
    raise errors.FuchsiaSdkError('Failed to add keys to Fuchsia ZBI: ' +
                                 result.output)
  os.chmod(initrd_path, 0o644)


def get_application_launch_command(arguments, testcase_path):
  """Prepare a command to run on the host to launch on the device."""
  # TODO(flowerhack): Implement this.
  return ''


def reset_state():
  """Reset the device to a clean state."""
  # TODO(flowerhack): Implement this.


def run_command(command_line, timeout):
  """Run the desired command on the device."""
  # TODO(flowerhack): Implement this.


def clear_testcase_directory():
  """Delete test cases stored on the device."""
  # TODO(flowerhack): Implement this.


def copy_testcase_to_device(testcase_path):
  """Copy a file to the device's test case directory."""
  # TODO(flowerhack): Implement this.



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions related to fuzz target entities."""
from google.cloud import ndb

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils


def get_fuzz_targets_for_target_jobs(target_jobs):
  """Return corresponding FuzzTargets for the given FuzzTargetJobs."""
  target_keys = [
      ndb.Key(data_types.FuzzTarget, t.fuzz_target_name) for t in target_jobs
  ]
  return ndb_utils.get_multi(target_keys)


def get_fuzz_target_jobs(fuzz_target_name=None,
                         engine=None,
                         job=None,
                         limit=None):
  """Return a Datastore query for fuzz target to job mappings."""
  query = data_types.FuzzTargetJob.query()

  if fuzz_target_name:
    query = query.filter(
        data_types.FuzzTargetJob.fuzz_target_name == fuzz_target_name)

  if job:
    query = query.filter(data_types.FuzzTargetJob.job == job)

  if engine:
    query = query.filter(data_types.FuzzTargetJob.engine == engine)

  if limit is not None:
    return query.iter(limit=limit)

  return ndb_utils.get_all_from_query(query)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Constants for ML RNN model."""

# Model name.
RNN_MODEL_NAME = 'rnn'

# Num of chars to be trained in one batch.
TRAINING_SEQLEN = 30

# Num of chars to be validated in one batch.
VALIDATION_SEQLEN = 1024

# Size of the alphabet that we work with.
ALPHA_SIZE = 256

# Num of sequences in a batch.
BATCH_SIZE = 100

# Number of epochs.
EPOCHS = 10

# Size of internal states in a neural cell.
HIDDEN_STATE_SIZE = 512

# Number of hidden layers.
HIDDEN_LAYER_SIZE = 3

# Learning rate.
LEARNING_RATE = 0.001

# Dropout probability (keep rate).
DROPOUT_PKEEP = 0.8

# Max num of models to keep during training.
MAX_TO_KEEP = 10

# Display training progress for every 50 batches.
DISPLAY_FREQ = 50

# Length of progress bar.
DISPLAY_LEN = 138

# Training script name.
TRAINING_SCRIPT_NAME = 'train.py'

# Generation script name.
GENERATION_SCRIPT_NAME = 'generate.py'

# Model files suffix.
MODEL_DATA_SUFFIX = '.data-00000-of-00001'
MODEL_INDEX_SUFFIX = '.index'

# Model parameter arguments.
BATCH_SIZE_ARGUMENT_PREFIX = '--batch-size='
HIDDEN_STATE_ARGUMENT_PREFIX = '--hidden-state-size='
HIDDEN_LAYER_ARGUMENT_PREFIX = '--hidden-layer-size='

# Training script arguments.
INPUT_DIR_ARGUMENT_PREFIX = '--input-dir='
MODEL_DIR_ARGUMENT_PREFIX = '--model-dir='
LOG_DIR_ARGUMENT_PREFIX = '--log-dir='

# Generation script arguments.
OUTPUT_DIR_ARGUMENT_PREFIX = '--output-dir='
MODEL_PATH_ARGUMENT_PREFIX = '--model-path='
GENERATION_COUNT_ARGUMENT_PREFIX = '--count='


class ExitCode(object):
  """Exit code for training and generation."""
  SUCCESS = 0
  INVALID_PATH = 1
  CORPUS_TOO_SMALL = 2
  TENSORFLOW_ERROR = 3



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Update task for updating source and tests."""

import datetime
import os
import platform
import sys
import time
import zipfile

from clusterfuzz._internal.base import dates
from clusterfuzz._internal.base import persistent_cache
from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.init_scripts import android as android_init
from clusterfuzz._internal.bot.init_scripts import chromeos as chromeos_init
from clusterfuzz._internal.bot.init_scripts import fuchsia as fuchsia_init
from clusterfuzz._internal.bot.init_scripts import linux as linux_init
from clusterfuzz._internal.bot.init_scripts import mac as mac_init
from clusterfuzz._internal.bot.init_scripts import windows as windows_init
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.metrics import monitoring_metrics
from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import process_handler
from clusterfuzz._internal.system import shell

TESTS_LAST_UPDATE_KEY = 'tests_last_update'
TESTS_UPDATE_INTERVAL_DAYS = 1

MANIFEST_FILENAME = 'clusterfuzz-source.manifest'
if sys.version_info.major == 3:
  MANIFEST_FILENAME += '.3'


def _rename_dll_for_update(absolute_filepath):
  """Rename a DLL to allow for updates."""
  backup_filepath = absolute_filepath + '.bak.' + str(int(time.time()))
  os.rename(absolute_filepath, backup_filepath)


def _platform_deployment_filename():
  """Return the platform deployment filename."""
  platform_mappings = {
      'Linux': 'linux',
      'Windows': 'windows',
      'Darwin': 'macos'
  }

  base_filename = platform_mappings[platform.system()]
  if sys.version_info.major == 3:
    base_filename += '-3'

  return base_filename + '.zip'


def _deployment_file_url(filename):
  """Helper to return deployment file url."""
  deployment_bucket = local_config.ProjectConfig().get('deployment.bucket')
  if not deployment_bucket:
    return None

  return 'gs://{bucket}/{name}'.format(bucket=deployment_bucket, name=filename)


def get_source_url():
  """Return the source URL."""
  return _deployment_file_url(_platform_deployment_filename())


def get_source_manifest_url():
  """Return the source manifest URL."""
  return _deployment_file_url(MANIFEST_FILENAME)


def clear_old_files(directory, extracted_file_set):
  """Remove files from the directory that isn't in the given file list."""
  for root_directory, _, filenames in shell.walk(directory):
    for filename in filenames:
      file_path = os.path.join(root_directory, filename)
      if file_path not in extracted_file_set:
        shell.remove_file(file_path)

  shell.remove_empty_directories(directory)


def clear_pyc_files(directory):
  """Recursively remove all .pyc files from the given directory"""
  for root_directory, _, filenames in shell.walk(directory):
    for filename in filenames:
      if not filename.endswith('.pyc'):
        continue

      file_path = os.path.join(root_directory, filename)
      shell.remove_file(file_path)


def track_revision():
  """Get the local revision and report as a metric."""
  revision = get_local_source_revision() or ''
  monitoring_metrics.BOT_COUNT.set(1, {'revision': revision})


def get_local_source_revision():
  """Return the local source revision."""
  return utils.current_source_version()


def get_remote_source_revision(source_manifest_url):
  """Get remote revision. We refactor this method out, so that we can mock
    it."""
  return storage.read_data(source_manifest_url).decode('utf-8').strip()


def get_newer_source_revision():
  """Returns the latest source revision if there is an update, or None if the
  current source is up to date."""
  if (environment.get_value('LOCAL_SRC') or
      environment.get_value('LOCAL_DEVELOPMENT')):
    logs.log('Using local source, skipping source code update.')
    return None

  root_directory = environment.get_value('ROOT_DIR')
  temp_directory = environment.get_value('BOT_TMPDIR')
  source_manifest_url = get_source_manifest_url()
  if (not get_source_url() or not source_manifest_url or not temp_directory or
      not root_directory):
    logs.log('Skipping source code update.')
    return None

  logs.log('Checking source code for updates.')
  try:
    source_version = get_remote_source_revision(source_manifest_url)
  except Exception:
    logs.log_error('Error occurred while checking source version.')
    return None

  local_source_version = get_local_source_revision()
  if not local_source_version:
    logs.log('No manifest found. Forcing an update.')
    return source_version

  logs.log('Local source code version: %s.' % local_source_version)
  logs.log('Remote source code version: %s.' % source_version)
  if local_source_version >= source_version:
    logs.log('Remote souce code <= local source code. No update.')
    # No source code update found. Source code is current, bail out.
    return None

  return source_version


def run_platform_init_scripts():
  """Run platform specific initialization scripts."""
  logs.log('Running platform initialization scripts.')

  plt = environment.platform()
  if environment.is_android_emulator():
    # Nothing to do here since emulator is not started yet.
    pass
  elif environment.is_android():
    android_init.run()
  elif plt == 'CHROMEOS':
    chromeos_init.run()
  elif plt == 'FUCHSIA':
    fuchsia_init.run()
  elif plt == 'LINUX':
    linux_init.run()
  elif plt == 'MAC':
    mac_init.run()
  elif plt == 'WINDOWS':
    windows_init.run()
  else:
    raise RuntimeError('Unsupported platform')

  logs.log('Completed running platform initialization scripts.')


def update_source_code():
  """Updates source code files with latest version from appengine."""
  process_handler.cleanup_stale_processes()
  shell.clear_temp_directory()

  root_directory = environment.get_value('ROOT_DIR')
  temp_directory = environment.get_value('BOT_TMPDIR')
  temp_archive = os.path.join(temp_directory, 'clusterfuzz-source.zip')
  try:
    storage.copy_file_from(get_source_url(), temp_archive)
  except Exception:
    logs.log_error('Could not retrieve source code archive from url.')
    return

  try:
    file_list = archive.get_file_list(temp_archive)
    zip_archive = zipfile.ZipFile(temp_archive, 'r')
  except Exception:
    logs.log_error('Bad zip file.')
    return

  src_directory = os.path.join(root_directory, 'src')
  output_directory = os.path.dirname(root_directory)
  error_occurred = False
  normalized_file_set = set()
  for filepath in file_list:
    filename = os.path.basename(filepath)

    # This file cannot be updated on the fly since it is running as server.
    if filename == 'adb':
      continue

    absolute_filepath = os.path.join(output_directory, filepath)
    if os.path.altsep:
      absolute_filepath = absolute_filepath.replace(os.path.altsep, os.path.sep)

    if os.path.realpath(absolute_filepath) != absolute_filepath:
      continue

    normalized_file_set.add(absolute_filepath)
    try:
      file_extension = os.path.splitext(filename)[1]

      # Remove any .so files first before overwriting, as they can be loaded
      # in the memory of existing processes. Overwriting them directly causes
      # segfaults in existing processes (e.g. run.py).
      if file_extension == '.so' and os.path.exists(absolute_filepath):
        os.remove(absolute_filepath)

      # On Windows, to update DLLs (and native .pyd extensions), we rename it
      # first so that we can install the new version.
      if (environment.platform() == 'WINDOWS' and
          file_extension in ['.dll', '.pyd'] and
          os.path.exists(absolute_filepath)):
        _rename_dll_for_update(absolute_filepath)
    except Exception:
      logs.log_error('Failed to remove or move %s before extracting new '
                     'version.' % absolute_filepath)

    try:
      extracted_path = zip_archive.extract(filepath, output_directory)
      external_attr = zip_archive.getinfo(filepath).external_attr
      mode = (external_attr >> 16) & 0o777
      mode |= 0o440
      os.chmod(extracted_path, mode)
    except:
      error_occurred = True
      logs.log_error(
          'Failed to extract file %s from source archive.' % filepath)

  zip_archive.close()

  if error_occurred:
    return

  clear_pyc_files(src_directory)
  clear_old_files(src_directory, normalized_file_set)

  local_manifest_path = os.path.join(root_directory,
                                     utils.LOCAL_SOURCE_MANIFEST)
  source_version = utils.read_data_from_file(
      local_manifest_path, eval_data=False).decode('utf-8').strip()
  logs.log('Source code updated to %s.' % source_version)


def update_tests_if_needed():
  """Updates layout tests every day."""
  data_directory = environment.get_value('FUZZ_DATA')
  error_occured = False
  expected_task_duration = 60 * 60  # 1 hour.
  retry_limit = environment.get_value('FAIL_RETRIES')
  temp_archive = os.path.join(data_directory, 'temp.zip')
  tests_url = environment.get_value('WEB_TESTS_URL')

  # Check if we have a valid tests url.
  if not tests_url:
    return

  # Layout test updates are usually disabled to speedup local testing.
  if environment.get_value('LOCAL_DEVELOPMENT'):
    return

  # |UPDATE_WEB_TESTS| env variable can be used to control our update behavior.
  if not environment.get_value('UPDATE_WEB_TESTS'):
    return

  last_modified_time = persistent_cache.get_value(
      TESTS_LAST_UPDATE_KEY, constructor=datetime.datetime.utcfromtimestamp)
  if (last_modified_time is not None and not dates.time_has_expired(
      last_modified_time, days=TESTS_UPDATE_INTERVAL_DAYS)):
    return

  logs.log('Updating layout tests.')
  tasks.track_task_start(
      tasks.Task('update_tests', '', ''), expected_task_duration)

  # Download and unpack the tests archive.
  for _ in range(retry_limit):
    try:
      shell.remove_directory(data_directory, recreate=True)
      storage.copy_file_from(tests_url, temp_archive)
      archive.unpack(temp_archive, data_directory, trusted=True)
      shell.remove_file(temp_archive)
      error_occured = False
      break
    except:
      logs.log_error(
          'Could not retrieve and unpack layout tests archive. Retrying.')
      error_occured = True

  if not error_occured:
    persistent_cache.set_value(
        TESTS_LAST_UPDATE_KEY, time.time(), persist_across_reboots=True)

  tasks.track_task_end()


def run():
  """Run update task."""
  # Since this code is particularly sensitive for bot stability, continue
  # execution but store the exception if anything goes wrong during one of these
  # steps.
  try:
    # Update heartbeat with current time.
    data_handler.update_heartbeat()

    # Check overall free disk space. If we are running too low, clear all
    # data directories like builds, fuzzers, data bundles, etc.
    shell.clear_data_directories_on_low_disk_space()

    # Download new layout tests once per day.
    update_tests_if_needed()
  except Exception:
    logs.log_error('Error occurred while running update task.')

  # Even if there is an exception in one of the other steps, we want to try to
  # update the source. If for some reason the source code update fails, it is
  # not necessary to run the init scripts.
  try:
    # If there is a newer revision, exit and let run.py update the source code.
    if get_newer_source_revision() is not None:
      if environment.is_trusted_host():
        from clusterfuzz._internal.bot.untrusted_runner import host
        host.update_worker()

      sys.exit(0)

    # Run platform specific initialization scripts.
    run_platform_init_scripts()
  except Exception:
    logs.log_error('Error occurred while running update task.')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for regression_task."""

import json
import os
import unittest

from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.bot.tasks import progression_task
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class WriteToBigqueryTest(unittest.TestCase):
  """Test _write_to_big_query."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.google_cloud_utils.big_query.write_range',
    ])

    self.testcase = data_types.Testcase(
        crash_type='type',
        crash_state='state',
        security_flag=True,
        fuzzer_name='libfuzzer',
        overridden_fuzzer_name='libfuzzer_pdf',
        job_type='some_job')

  def test_write(self):
    """Tests write."""
    progression_task._write_to_bigquery(self.testcase, 456, 789)  # pylint: disable=protected-access
    self.mock.write_range.assert_called_once_with(
        table_id='fixeds',
        testcase=self.testcase,
        range_name='fixed',
        start=456,
        end=789)


class TestcaseReproducesInRevisionTest(unittest.TestCase):
  """Test _testcase_reproduces_in_revision."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager.setup_regular_build',
        'clusterfuzz._internal.bot.testcase_manager.test_for_crash_with_retries',
        'clusterfuzz._internal.bot.testcase_manager.check_for_bad_build',
    ])

  def test_error_on_failed_setup(self):
    """Ensure that we throw an exception if we fail to set up a build."""
    os.environ['APP_NAME'] = 'app_name'
    # No need to implement a fake setup_regular_build. Since it's doing nothing,
    # we won't have the build directory properly set.
    with self.assertRaises(errors.BuildSetupError):
      progression_task._testcase_reproduces_in_revision(  # pylint: disable=protected-access
          None, '/tmp/blah', 'job_type', 1)


@test_utils.with_cloud_emulators('datastore')
class UpdateIssueMetadataTest(unittest.TestCase):
  """Test _update_issue_metadata."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.engine_common.find_fuzzer_path',
        'clusterfuzz._internal.bot.fuzzers.engine_common.get_all_issue_metadata',
    ])

    data_types.FuzzTarget(engine='libFuzzer', binary='fuzzer').put()
    self.mock.get_all_issue_metadata.return_value = {
        'issue_labels': 'label1',
        'issue_components': 'component1',
    }

    self.testcase = data_types.Testcase(
        overridden_fuzzer_name='libFuzzer_fuzzer')
    self.testcase.put()

  def test_update_issue_metadata_non_existent(self):
    """Test update issue metadata a testcase with no metadata."""
    progression_task._update_issue_metadata(self.testcase)  # pylint: disable=protected-access

    testcase = self.testcase.key.get()
    self.assertDictEqual({
        'issue_labels': 'label1',
        'issue_components': 'component1',
    }, json.loads(testcase.additional_metadata))

  def test_update_issue_metadata_replace(self):
    """Test update issue metadata a testcase with different metadata."""
    self.testcase.additional_metadata = json.dumps({
        'issue_labels': 'label1',
        'issue_components': 'component2',
    })
    progression_task._update_issue_metadata(self.testcase)  # pylint: disable=protected-access

    testcase = self.testcase.key.get()
    self.assertDictEqual({
        'issue_labels': 'label1',
        'issue_components': 'component1',
    }, json.loads(testcase.additional_metadata))

  def test_update_issue_metadata_same(self):
    """Test update issue metadata a testcase with the same metadata."""
    self.testcase.additional_metadata = json.dumps({
        'issue_labels': 'label1',
        'issue_components': 'component1',
    })
    self.testcase.put()

    self.testcase.crash_type = 'test'  # Should not be written.
    progression_task._update_issue_metadata(self.testcase)  # pylint: disable=protected-access

    testcase = self.testcase.key.get()
    self.assertDictEqual({
        'issue_labels': 'label1',
        'issue_components': 'component1',
    }, json.loads(testcase.additional_metadata))
    self.assertIsNone(testcase.crash_type)


@test_utils.with_cloud_emulators('datastore')
class StoreTestcaseForRegressionTesting(fake_filesystem_unittest.TestCase):
  """Test _store_testcase_for_regression_testing."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.google_cloud_utils.storage.copy_file_to',
    ])

    os.environ['CORPUS_BUCKET'] = 'corpus'

    fuzz_target = data_types.FuzzTarget(id='libFuzzer_test_project_test_fuzzer')
    fuzz_target.binary = 'test_fuzzer'
    fuzz_target.project = 'test_project'
    fuzz_target.engine = 'libFuzzer'
    fuzz_target.put()

    self.testcase = data_types.Testcase()
    self.testcase.fuzzer_name = 'libFuzzer'
    self.testcase.overridden_fuzzer_name = 'libFuzzer_test_project_test_fuzzer'
    self.testcase.job_type = 'job'
    self.testcase.bug_information = '123'
    self.testcase.open = False
    self.testcase.put()

    self.testcase_file_path = '/testcase'
    self.fs.create_file(self.testcase_file_path, contents='A')

  def test_open_testcase(self):
    """Test that an open testcase is not stored for regression testing."""
    self.testcase.open = True
    self.testcase.put()

    progression_task._store_testcase_for_regression_testing(  # pylint: disable=protected-access
        self.testcase, self.testcase_file_path)
    self.assertEqual(0, self.mock.copy_file_to.call_count)

  def test_testcase_with_no_issue(self):
    """Test that a testcase with no associated issue is not stored for
    regression testing."""
    self.testcase.bug_information = ''
    self.testcase.put()

    progression_task._store_testcase_for_regression_testing(  # pylint: disable=protected-access
        self.testcase, self.testcase_file_path)
    self.assertEqual(0, self.mock.copy_file_to.call_count)

  def test_testcase_with_no_fuzz_target(self):
    """Test that a testcase with no associated fuzz target is not stored for
    regression testing."""
    self.testcase.overridden_fuzzer_name = 'libFuzzer_not_exist'
    self.testcase.put()

    progression_task._store_testcase_for_regression_testing(  # pylint: disable=protected-access
        self.testcase, self.testcase_file_path)
    self.assertEqual(0, self.mock.copy_file_to.call_count)

  def test_testcase_stored(self):
    """Test that a testcase is stored for regression testing."""
    progression_task._store_testcase_for_regression_testing(  # pylint: disable=protected-access
        self.testcase, self.testcase_file_path)
    self.mock.copy_file_to.assert_called_with(
        '/testcase',
        'gs://corpus/libFuzzer/test_project_test_fuzzer_regressions/'
        '6dcd4ce23d88e2ee9568ba546c007c63d9131c1b')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Date or time helper functions."""

import time

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.system import environment


def initialize_timezone_from_environment():
  """Initializes timezone for date functions based on environment."""
  plt = environment.platform()
  if plt == 'WINDOWS':
    return

  # Only available on Unix platforms.
  time.tzset()


def time_has_expired(timestamp,
                     compare_to=None,
                     days=0,
                     hours=0,
                     minutes=0,
                     seconds=0):
  """Checks to see if a timestamp is older than another by a certain amount."""
  if compare_to is None:
    compare_to = utils.utcnow()

  total_time = days * 3600 * 24 + hours * 3600 + minutes * 60 + seconds
  return (compare_to - timestamp).total_seconds() > total_time



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Storage tests."""
import datetime
import os
import unittest

from pyfakefs import fake_filesystem_unittest
import six

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.tests.test_libs import test_utils


class LifecycleConfigTest(unittest.TestCase):
  """Test lifecycle config generation."""

  def test_empty(self):
    """Test generation of an empty config."""
    config = storage.generate_life_cycle_config('')
    expected_config = {
        'rule': [{
            'action': {
                'type': '',
            },
            'condition': {},
        }],
    }
    self.assertEqual(config, expected_config)

  def test_delete_with_age(self):
    """Test generation of a config for deletion with age condition."""
    config = storage.generate_life_cycle_config('Delete', age=7)
    expected_config = {
        'rule': [{
            'action': {
                'type': 'Delete',
            },
            'condition': {
                'age': 7,
            },
        }],
    }
    self.assertEqual(config, expected_config)

  def test_delete_with_age_and_new_versions(self):
    """Test generation of a config for deletion with age and numNewerVersions
    condition."""
    config = storage.generate_life_cycle_config(
        'Delete', age=30, num_newer_versions=10)
    expected_config = {
        'rule': [{
            'action': {
                'type': 'Delete',
            },
            'condition': {
                'age': 30,
                'numNewerVersions': 10,
            },
        }],
    }
    self.assertEqual(config, expected_config)


class FileSystemProviderTests(fake_filesystem_unittest.TestCase):
  """Tests for FileSystemProvider."""

  def setUp(self):
    self.provider = storage.FileSystemProvider('/local')
    test_utils.set_up_pyfakefs(self)

  def test_create_bucket(self):
    """Test create_bucket."""
    self.provider.create_bucket('test-bucket', None, None)
    self.assertTrue(os.path.isdir('/local/test-bucket'))

  def test_get_bucket(self):
    """Test get_bucket."""
    self.fs.create_dir('/local/test-bucket')
    self.assertDictEqual({
        'name': 'test-bucket',
    }, self.provider.get_bucket('test-bucket'))

  def test_list_blobs(self):
    """Test list_blobs."""
    mtime = datetime.datetime(2019, 1, 1)
    mtime_seconds = utils.utc_datetime_to_timestamp(mtime)

    self.fs.create_file(
        '/local/test-bucket/objects/a', st_size=11).st_mtime = mtime_seconds
    self.fs.create_file(
        '/local/test-bucket/objects/b/c', st_size=22).st_mtime = mtime_seconds
    self.fs.create_file(
        '/local/test-bucket/objects/b/d/e', st_size=33).st_mtime = mtime_seconds
    self.fs.create_file(
        '/local/test-bucket/objects/f', st_size=44).st_mtime = mtime_seconds
    self.fs.create_file(
        '/local/test-bucket/metadata/b/c',
        contents='{"key":"value"}').st_mtime = mtime_seconds

    result = list(self.provider.list_blobs('gs://test-bucket'))
    six.assertCountEqual(self, [{
        'bucket': 'test-bucket',
        'name': 'a',
        'updated': mtime,
        'size': 11,
        'metadata': {}
    }, {
        'bucket': 'test-bucket',
        'name': 'f',
        'updated': mtime,
        'size': 44,
        'metadata': {}
    }, {
        'bucket': 'test-bucket',
        'name': 'b/c',
        'updated': mtime,
        'size': 22,
        'metadata': {
            'key': 'value'
        }
    }, {
        'bucket': 'test-bucket',
        'name': 'b/d/e',
        'updated': mtime,
        'size': 33,
        'metadata': {}
    }], result)

    result = list(self.provider.list_blobs('gs://test-bucket/b'))
    six.assertCountEqual(self, [{
        'bucket': 'test-bucket',
        'name': 'b/c',
        'updated': mtime,
        'size': 22,
        'metadata': {
            'key': 'value'
        }
    }, {
        'bucket': 'test-bucket',
        'name': 'b/d/e',
        'updated': mtime,
        'size': 33,
        'metadata': {}
    }], result)

    result = list(self.provider.list_blobs('gs://test-bucket/b/d'))
    six.assertCountEqual(self, [{
        'bucket': 'test-bucket',
        'name': 'b/d/e',
        'updated': mtime,
        'size': 33,
        'metadata': {}
    }], result)

    result = list(
        self.provider.list_blobs('gs://test-bucket/', recursive=False))
    six.assertCountEqual(self, [{
        'bucket': 'test-bucket',
        'name': 'a',
        'updated': mtime,
        'size': 11,
        'metadata': {}
    }, {
        'bucket': 'test-bucket',
        'name': 'f',
        'updated': mtime,
        'size': 44,
        'metadata': {}
    }, {
        'bucket': 'test-bucket',
        'name': 'b',
    }], result)

  def test_copy_file_from(self):
    """Test copy_file_from."""
    self.fs.create_file('/local/test-bucket/objects/a', contents='a')
    self.provider.copy_file_from('gs://test-bucket/a', '/a')
    with open('/a') as f:
      self.assertEqual('a', f.read())

  def test_copy_file_to(self):
    """Test copy_file_to."""
    self.fs.create_file('/a', contents='a')
    self.fs.create_dir('/local/test-bucket')

    self.provider.copy_file_to(
        '/a', 'gs://test-bucket/subdir/a', metadata={'key': 'value'})
    with open('/local/test-bucket/objects/subdir/a') as f:
      self.assertEqual('a', f.read())

    with open('/local/test-bucket/metadata/subdir/a') as f:
      self.assertEqual('{"key": "value"}', f.read())

  def test_copy_blob(self):
    """Test copy_blob."""
    self.fs.create_file('/local/test-bucket/objects/a', contents='a')

    self.provider.copy_blob('gs://test-bucket/a', 'gs://test-bucket/copy/a')
    with open('/local/test-bucket/objects/copy/a') as f:
      self.assertEqual('a', f.read())

  def test_read_data(self):
    """Test copy_blob."""
    self.fs.create_file('/local/test-bucket/objects/a', contents='a')
    self.assertEqual(b'a', self.provider.read_data('gs://test-bucket/a'))

  def test_write_data(self):
    """Test copy_blob."""
    self.fs.create_dir('/local/test-bucket')
    self.provider.write_data(
        b'a', 'gs://test-bucket/subdir/a', metadata={'key': 'value'})
    with open('/local/test-bucket/objects/subdir/a') as f:
      self.assertEqual('a', f.read())

    with open('/local/test-bucket/metadata/subdir/a') as f:
      self.assertEqual('{"key": "value"}', f.read())

    self.provider.write_data('b', 'gs://test-bucket/subdir/b')
    with open('/local/test-bucket/objects/subdir/b') as f:
      self.assertEqual('b', f.read())

  def test_get(self):
    """Test get."""
    mtime = datetime.datetime(2019, 1, 1)
    mtime_seconds = utils.utc_datetime_to_timestamp(mtime)

    self.fs.create_file(
        '/local/test-bucket/objects/a', contents='a').st_mtime = mtime_seconds
    self.fs.create_file(
        '/local/test-bucket/metadata/a', contents='{"key": "value"}')

    self.assertDictEqual({
        'bucket': 'test-bucket',
        'name': 'a',
        'size': 1,
        'updated': mtime,
        'metadata': {
            'key': 'value'
        },
    }, self.provider.get('gs://test-bucket/a'))

  def test_delete(self):
    """Test get."""
    self.fs.create_file('/local/test-bucket/objects/a', contents='a')
    self.fs.create_file(
        '/local/test-bucket/metadata/a', contents='{"key": "value"}')

    self.provider.delete('gs://test-bucket/a')
    self.assertFalse(os.path.exists('/local/test-bucket/objects/a'))
    self.assertFalse(os.path.exists('/local/test-bucket/metadata/a'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Console input functions for the reproduce tool."""


def get_string(prompt):
  """Prompt the user for a string from console input."""
  return input(prompt + ': ')


def get_boolean(prompt):
  """Return a boolean representing a yes/no answer to a prompt."""
  result = get_string(prompt + ' (Y/n)')
  return result.lower() == 'y'



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for helping in crash comparison."""

# pylint: disable=consider-using-enumerate


def _levenshtein_distance(string_1, string_2):
  """"Levenshtein_distance calculation: Iterative with two matrix rows,
  based on Wikipedia article and code by Christopher P. Matthews."""
  if string_1 == string_2:
    return 0
  if not string_1:
    return len(string_2)
  if not string_2:
    return len(string_1)

  v0 = list(range(len(string_2) + 1))
  v1 = [None] * (len(string_2) + 1)

  for i in range(len(string_1)):
    v1[0] = i + 1

    for j in range(len(string_2)):
      cost = 0 if string_1[i] == string_2[j] else 1
      v1[j + 1] = min(v1[j] + 1, v0[j + 1] + 1, v0[j] + cost)

    for j in range(len(v0)):
      v0[j] = v1[j]

  return v1[len(string_2)]


def _similarity_ratio(string_1, string_2):
  """Return a ratio on how similar two strings are."""
  length_sum = len(string_1) + len(string_2)
  if length_sum == 0:
    return 1.0

  return (length_sum - _levenshtein_distance(string_1, string_2)) / (
      1.0 * length_sum)


def longest_common_subsequence(first_frames, second_frames):
  """Count number of frames which are the same (taking into account order)."""
  first_len = len(first_frames)
  second_len = len(second_frames)

  solution = [[0 for _ in range(second_len + 1)] for _ in range(first_len + 1)]

  for i in range(1, first_len + 1):
    for j in range(1, second_len + 1):
      if first_frames[i - 1] == second_frames[j - 1]:
        solution[i][j] = solution[i - 1][j - 1] + 1
      else:
        solution[i][j] = max(solution[i - 1][j], solution[i][j - 1])

  return solution[first_len][second_len]


class CrashComparer(object):
  """Compares two crash results."""
  COMPARE_THRESHOLD = 0.8
  SAME_FRAMES_THRESHOLD = 2

  def __init__(self, crash_state_1, crash_state_2, compare_threshold=None):
    self.crash_state_1 = crash_state_1
    self.crash_state_2 = crash_state_2
    self.compare_threshold = compare_threshold or self.COMPARE_THRESHOLD

  def is_similar(self):
    """Return a bool for whether the two crash results are similar."""
    # If one of the crash state is empty, it can't match anything.
    if not self.crash_state_1 or not self.crash_state_2:
      return False

    # Optimization: Do a == check first before others.
    if self.crash_state_1 == self.crash_state_2:
      return True

    # If there is a fuzzer hash string in state, then rely on exact comparison.
    # Since we failed the check above, our hashes don't match.
    if 'FuzzerHash=' in self.crash_state_1:
      return False

    # TODO(aarya): Improve this algorithm and leverage other parts of
    # stacktrace.
    crash_state_lines_1 = self.crash_state_1.splitlines()
    crash_state_lines_2 = self.crash_state_2.splitlines()

    if (longest_common_subsequence(crash_state_lines_1, crash_state_lines_2) >=
        self.SAME_FRAMES_THRESHOLD):
      return True

    lines_compared = 0
    similarity_ratio_sum = 0.0
    for i in range(len(crash_state_lines_1)):
      if i >= len(crash_state_lines_2):
        break

      similarity_ratio = _similarity_ratio(crash_state_lines_1[i],
                                           crash_state_lines_2[i])
      lines_compared += 1
      similarity_ratio_sum += similarity_ratio

    similarity_ratio_average = similarity_ratio_sum / lines_compared
    return similarity_ratio_average > self.compare_threshold



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for handling Fuchsia hosts."""

from collections import defaultdict
import json
import os
import re
import subprocess

from .process import Process


class Host(object):
  """Represents a local system with a build of Fuchsia.

    This class abstracts the details of various repository, tool, and build
    paths, as well as details about the host architecture and platform.

    Attributes:
      fuzzers:   The fuzzer binaries available in the current Fuchsia image
      build_dir: The build output directory, if present.
  """

  # Convenience file descriptor for silencing subprocess output
  DEVNULL = open(os.devnull, 'w')

  class ConfigError(RuntimeError):
    """Indicates the host is not configured for building Fuchsia."""

  @classmethod
  def from_build(cls):
    """Uses a local build directory to configure a Host object from it."""
    host = Host()
    host.set_build_dir(host.find_build_dir())
    return host

  @classmethod
  def from_dir(cls, this_dir):
    host = Host()
    host.set_build_dir(this_dir)
    return host

  @classmethod
  def join(cls, *segments):
    """Creates a source tree path."""
    fuchsia = os.getenv('FUCHSIA_DIR')
    if not fuchsia:
      raise Host.ConfigError('Unable to find FUCHSIA_DIR; have you `fx set`?')
    return os.path.join(fuchsia, *segments)

  def __init__(self):
    self._ids = []
    self._llvm_symbolizer = None
    self._platform = 'mac-x64' if os.uname()[0] == 'Darwin' else 'linux-x64'
    self._symbolizer_exec = None
    self._zxtools = None
    self.fuzzers = []
    self.build_dir = None

  @classmethod
  def find_build_dir(cls):
    """Examines the source tree to locate a build directory."""
    build_dir = Host.join('.fx-build-dir')
    if not os.path.exists(build_dir):
      raise Host.ConfigError('Unable to find .fx-build-dir; have you `fx set`?')
    with open(build_dir, 'r') as f:
      return Host.join(f.read().strip())

  def add_build_ids(self, build_ids):
    """Sets the build IDs used to symbolize logs."""
    if os.path.exists(build_ids):
      self._ids.append(build_ids)

  def set_zxtools(self, zxtools):
    """Sets the location of the Zircon host tools directory."""
    if not os.path.isdir(zxtools):
      raise Host.ConfigError('Unable to find Zircon host tools.')
    self._zxtools = zxtools

  def set_symbolizer(self, executable, symbolizer):
    """Sets the paths to both the wrapper and LLVM symbolizers."""
    if not os.path.exists(executable) or not os.access(executable, os.X_OK):
      raise Host.ConfigError('Invalid symbolize binary: ' + executable)
    if not os.path.exists(symbolizer) or not os.access(symbolizer, os.X_OK):
      raise Host.ConfigError('Invalid LLVM symbolizer: ' + symbolizer)
    self._symbolizer_exec = executable
    self._llvm_symbolizer = symbolizer

  def set_fuzzers_json(self, json_file):
    """Sets the path to the build file with fuzzer metadata."""
    if not os.path.exists(json_file):
      raise Host.ConfigError('Unable to find list of fuzzers.')
    self.fuzzers = []
    with open(json_file) as f:
      fuzz_specs = json.load(f)
    by_label = defaultdict(dict)
    for fuzz_spec in fuzz_specs:
      # Try v2 metadata first.
      label = fuzz_spec.get('label')
      if label:
        by_label[label].update(fuzz_spec)
        continue
      # Fallback to v1 metadata.
      pkg = fuzz_spec['fuzzers_package']
      self.fuzzers += [(pkg, tgt) for tgt in fuzz_spec['fuzzers']]
    self.fuzzers += [(fuzz_spec['package'], fuzz_spec['fuzzer'])
                     for fuzz_spec in by_label.values()]

  def set_build_dir(self, build_dir):
    """Configure the host using data from a build directory."""
    self.set_zxtools(Host.join(build_dir + '.zircon', 'tools'))
    clang_dir = os.path.join('buildtools', self._platform, 'clang')
    self.set_symbolizer(
        Host.join('zircon', 'prebuilt', 'downloads', 'symbolize',
                  self._platform, 'symbolize'),
        Host.join(clang_dir, 'bin', 'llvm-symbolizer'))
    self.add_build_ids(Host.join('prebuilt_build_ids'))
    self.add_build_ids(Host.join(clang_dir, 'lib', 'debug', '.build-id'))
    self.add_build_ids(Host.join(build_dir, '.build-id'))
    self.add_build_ids(Host.join(build_dir + '.zircon', '.build-id'))
    json_file = Host.join(build_dir, 'fuzzers.json')
    # fuzzers.json isn't emitted in release builds
    if os.path.exists(json_file):
      self.set_fuzzers_json(json_file)
    self.build_dir = build_dir

  def create_process(self, args, **kwargs):
    return Process(args, **kwargs)

  def zircon_tool(self, cmd, logfile=None):
    """Executes a tool found in the ZIRCON_BUILD_DIR."""
    if not self._zxtools:
      raise Host.ConfigError('Zircon host tools unavailable.')
    if not os.path.isabs(cmd[0]):
      cmd[0] = os.path.join(self._zxtools, cmd[0])
    if not os.path.exists(cmd[0]):
      raise Host.ConfigError('Unable to find Zircon host tool: ' + cmd[0])
    p = self.create_process(cmd)
    if logfile:
      p.stdout = logfile
      p.stderr = subprocess.STDOUT
      return p.popen()
    p.stderr = Host.DEVNULL
    return p.check_output().strip()

  def killall(self, process):
    """ Invokes killall on the process name."""
    p = self.create_process(
        ['killall', process], stdout=Host.DEVNULL, stderr=Host.DEVNULL)
    p.call()

  def symbolize(self, raw):
    """Symbolizes backtraces in a log file using the current build.

        Attributes:
            raw: A list of unsymbolized lines.

        Returns:
            A list of symbolized lines.
        """
    if not self._symbolizer_exec:
      raise Host.ConfigError('Symbolizer executable not set.')
    if not self._ids:
      raise Host.ConfigError('Build IDs not set.')
    if not self._llvm_symbolizer:
      raise Host.ConfigError('LLVM symbolizer not set.')

    # Symbolize
    args = [self._symbolizer_exec, '-llvm-symbolizer', self._llvm_symbolizer]
    for build_id_dir in self._ids:
      args += ['-build-id-dir', build_id_dir]
    p = self.create_process(args)
    p.stdin = subprocess.PIPE
    p.stdout = subprocess.PIPE
    out, _ = p.popen().communicate(raw)
    processed_out = re.sub(br'[0-9\[\]\.]*\[klog\] INFO: ', b'', out)
    return processed_out.split(b'\n')

  def notify_user(self, title, body):
    """Displays a message to the user in a platform-specific way"""
    args = ['which', 'notify-send']
    p = self.create_process(args, stdout=Host.DEVNULL, stderr=Host.DEVNULL)

    if self._platform == 'mac-x64':
      args = [
          'osascript', '-e',
          'display notification "' + body + '" with title "' + title + '"'
      ]
    elif p.call() == 0:
      args = ['notify-send', title, body]
    else:
      args = ['wall', title + '; ' + body]
    return self.create_process(
        args, stdout=Host.DEVNULL, stderr=Host.DEVNULL).call()

  def snapshot(self):
    integration = Host.join('integration')
    if not os.path.isdir(integration):
      raise Host.ConfigError('Missing integration repo.')
    p = self.create_process(
        ['git', 'rev-parse', 'HEAD'], stderr=Host.DEVNULL, cwd=integration)
    return p.check_output().strip()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzing engine initialization."""

from clusterfuzz._internal.bot.fuzzers.afl import engine as afl_engine
from clusterfuzz._internal.bot.fuzzers.blackbox import engine as blackbox_engine
from clusterfuzz._internal.bot.fuzzers.googlefuzztest import \
    engine as gft_engine
from clusterfuzz._internal.bot.fuzzers.honggfuzz import \
    engine as honggfuzz_engine
from clusterfuzz._internal.bot.fuzzers.libFuzzer import \
    engine as libFuzzer_engine
from clusterfuzz._internal.bot.fuzzers.syzkaller import \
    engine as syzkaller_engine
from clusterfuzz.fuzz import engine


def run():
  """Initialise builtin fuzzing engines."""
  engine.register('afl', afl_engine.AFLEngine)
  engine.register('blackbox', blackbox_engine.BlackboxEngine)
  engine.register('googlefuzztest', gft_engine.GoogleFuzzTestEngine)
  engine.register('honggfuzz', honggfuzz_engine.HonggfuzzEngine)
  engine.register('libFuzzer', libFuzzer_engine.LibFuzzerEngine)
  engine.register('syzkaller', syzkaller_engine.SyzkallerEngine)
